#cython: language_level=3
 

__OBIDMS_COLUMN_CLASS__ = {}


from ..capi.obitypes cimport name_data_type, \
                             obitype_t, \
                             obiversion_t, \
                             OBI_QUAL, \
                             OBI_STR

from ..capi.obidms cimport obi_import_column

from ..capi.obidmscolumn cimport OBIDMS_column_header_p, \
                                 obi_close_column, \
                                 obi_get_elements_names, \
                                 obi_column_formatted_infos, \
                                 obi_column_write_comments
                                     
from ..capi.obiutils cimport obi_format_date
 
from ..capi.obiview cimport obi_view_add_column, \
                            obi_view_get_pointer_on_column_in_view, \
                            Obiview_p, \
                            NUC_SEQUENCE_COLUMN, \
                            QUALITY_COLUMN, \
                            REVERSE_SEQUENCE_COLUMN, \
                            REVERSE_QUALITY_COLUMN


from ..object cimport OBIDeactivatedInstanceError

from obitools3.utils cimport tobytes, \
                             bytes2str, \
                             str2bytes, \
                             str2bytes_object, \
                             bytes2str_object, \
                             clean_empty_values_from_object
 
from obitools3.dms.column import typed_column
 
from libc.stdlib cimport free
from libc.string cimport strcpy
  
import importlib
import inspect
import pkgutil
import json


cdef class Column(OBIWrapper) :
    '''
    The obitools3.dms.column.Column class wraps a C instance of a column in the context of a View 
    '''

    cdef inline OBIDMS_column_p pointer(self) :
        return <OBIDMS_column_p>(<OBIDMS_column_p*>(self._pointer))[0]


    @staticmethod
    cdef type get_column_class(obitype_t obitype, bint multi_elts, bint tuples):
        '''
        Internal function returning the python class representing 
        a column for a given obitype.
        '''
        return __OBIDMS_COLUMN_CLASS__[(obitype, multi_elts, tuples)][0]


    @staticmethod
    cdef type get_python_type(obitype_t obitype, bint multi_elts):  # TODO
        '''
        Internal function returning the python type representing 
        an instance for a given obitype.
        '''
        return __OBIDMS_COLUMN_CLASS__[(obitype, multi_elts)][1]


    @staticmethod
    def import_column(object dms_1, object dms_2, object column_name, obiversion_t version_number):
        cdef obiversion_t new_version
        new_version = obi_import_column(tobytes(dms_1), tobytes(dms_2), tobytes(column_name), version_number)
        if new_version < 0 :
            raise Exception("Error importing a column")
        return new_version


    @staticmethod
    def new_column(View      view,
                   object    column_name,
                   obitype_t data_type,
                   index_t   nb_elements_per_line=1,
                   list      elements_names=None,
                   bint      dict_column=False,
                   bint      tuples=False,
                   bint      to_eval=False,
                   object    associated_column_name=b"",
                   int       associated_column_version=-1,
                   object    comments={},
                   object    alias=b""):
        # TODO indexer_name?
        
        cdef Column column
        cdef bytes  column_name_b = tobytes(column_name)
        cdef bytes  alias_b = tobytes(alias)
        cdef bytes  comments_b = str2bytes(json.dumps(bytes2str_object(comments)))
        cdef bytes  associated_column_name_b = tobytes(associated_column_name)
        cdef list   elements_names_s
        cdef bytes  elements_names_b
        cdef char*  elements_names_p
        cdef object elt_name
        cdef str    elt_name_s
        
        if not view.active() :
            raise OBIDeactivatedInstanceError()
 
        if alias_b == b"" :
            alias_b = column_name_b
        
        if elements_names is not None:
            elements_names_s = []
            for elt_name in elements_names :
                if type(elt_name) != bytes and type(elt_name) != str :
                    elements_names_s.append(str(elt_name))
                else :
                    elements_names_s.append(elt_name)
            elements_names_b = b';'.join([tobytes(x) for x in elements_names_s])
            elements_names_p = elements_names_b
        else:
            elements_names_p = NULL
        
        if column_name_b == b"SAMPLE" or column_name_b == b"sample":
        # force str type
            data_type = OBI_STR
        
        if data_type == OBI_QUAL:
            if associated_column_name_b == b"":
                if column_name == QUALITY_COLUMN:
                    if NUC_SEQUENCE_COLUMN not in view:
                         raise RuntimeError("Cannot create column %s in view %s: trying to create quality column but no NUC_SEQ column to associate it with in the view" % (bytes2str(column_name_b),
                                                                               bytes2str(view.name)))
                    associated_column_name_b = NUC_SEQUENCE_COLUMN
                    associated_column_version = view[NUC_SEQUENCE_COLUMN].version                    
                elif column_name == REVERSE_QUALITY_COLUMN:
                    if REVERSE_SEQUENCE_COLUMN not in view:
                         raise RuntimeError("Cannot create column %s in view %s: trying to create reverse quality column but no REVERSE_SEQUENCE column to associate it with in the view" % (bytes2str(column_name_b),
                                                                               bytes2str(view.name)))
                    associated_column_name_b = REVERSE_SEQUENCE_COLUMN
                    associated_column_version = view[REVERSE_SEQUENCE_COLUMN].version
                    
        
        if (obi_view_add_column(view                      = view.pointer(),
                                column_name               = column_name_b,
                                version_number            = -1,
                                alias                     = alias_b,
                                data_type                 = <obitype_t>data_type,
                                nb_lines                  = len(view),
                                nb_elements_per_line      = nb_elements_per_line,
                                elements_names            = elements_names_p,
                                elt_names_formatted       = False,
                                dict_column               = dict_column,
                                tuples                    = tuples,
                                to_eval                   = to_eval,
                                indexer_name              = NULL,
                                associated_column_name    = associated_column_name_b,
                                associated_column_version = associated_column_version,
                                comments                  = comments_b,
                                create                    = True)<0):
            raise RuntimeError("Cannot create column %s in view %s" % (bytes2str(column_name_b),
                                                                       bytes2str(view.name)))
        
        column = Column.open(view, alias_b)
        
        # Automatically associate nuc sequence column to quality column if necessary
        if data_type == OBI_QUAL:
            if column_name == QUALITY_COLUMN:
                view[NUC_SEQUENCE_COLUMN].associated_column_name = column.name
                view[NUC_SEQUENCE_COLUMN].associated_column_version = column.version
            elif column_name == REVERSE_QUALITY_COLUMN:
                view[REVERSE_SEQUENCE_COLUMN].associated_column_name = column.name
                view[REVERSE_SEQUENCE_COLUMN].associated_column_version = column.version
        
        return column
 
 
    @staticmethod
    def open(View view,
             object column_name):
        cdef bytes            column_name_b = tobytes(column_name)
        cdef OBIDMS_column_p* column_pp
        cdef OBIDMS_column_p  column_p
        cdef Column           column
        cdef obitype_t        column_type
        cdef type             column_class
        
        if not view.active() :
            raise OBIDeactivatedInstanceError()
        
        column_pp = obi_view_get_pointer_on_column_in_view(view.pointer(), 
                                                           column_name_b)
        
        if column_pp == NULL:
            raise KeyError("Cannot access to column %s in view %s" % (
                    bytes2str(column_name_b),
                    bytes2str(view.name)
                ))
        
        column_p = column_pp[0]
        column_type = column_p.header.returned_data_type
        column_class = Column.get_column_class(column_type, (column_p.header.nb_elements_per_line > 1 or column_p.header.dict_column == True), column_p.header.tuples)
        column = OBIWrapper.new_wrapper(column_class, column_pp)
        
        column._view = view
        column._alias = column_name_b
        column._elements_names = column.read_elements_names()
        view.register(column)

        return column

    
    @OBIWrapper.checkIsActive
    def add_to_view(self,
                    View   view,
                    object column_name=None) :
        
        cdef bytes alias   
        cdef OBIDMS_column_p column_p = self.pointer()
        
        if not view.active() :
            raise OBIDeactivatedInstanceError()
        
        if (column_name is None):
            alias = self._alias
        else:
            alias = tobytes(column_name)
                       
        if (obi_view_add_column(view                      = view.pointer(),
                                column_name               = column_p.header.name,
                                version_number            = column_p.header.version,
                                alias                     = alias,
                                data_type                 = <obitype_t>0,
                                nb_lines                  = -1,
                                nb_elements_per_line      = -1,
                                elements_names            = NULL,
                                elt_names_formatted       = False,
                                dict_column               = False,
                                tuples                    = False,
                                to_eval                   = False,
                                indexer_name              = NULL,
                                associated_column_name    = NULL,
                                associated_column_version = -1,
                                comments                  = NULL,
                                create                    = False) < 0):
            raise RuntimeError("Cannot insert column %s (%s@%d) into view %s" %
                               ( bytes2str(alias),
                                 bytes2str(column_p.header.name),
                                 column_p.header.version,
                                 bytes2str(view.name)
                               ))
            
        view.register(self)
        
    
    @OBIWrapper.checkIsActive
    def __len__(self):
        '''
        implements the len() function for the Column class
         
            @rtype: `int`
        '''
        return self.lines_used
     
    
    @OBIWrapper.checkIsActive 
    def __sizeof__(self):
        '''
        returns the size of the C object wrapped by the Column instance
        '''
        cdef OBIDMS_column_header_p header = self.pointer().header
        return header.header_size + header.data_size
     
    
    @OBIWrapper.checkIsActive 
    def __iter__(self):
        cdef index_t line_nb
        for line_nb in range(self.lines_used):
            yield self.get_line(line_nb)

    
    # TODO check time efficiency with and without
    @OBIWrapper.checkIsActive
    def __setitem__(self, index_t line_nb, object value):
        self.set_line(line_nb, value)


    @OBIWrapper.checkIsActive
    def __getitem__(self, index_t line_nb):
        return self.get_line(line_nb)
  
    
    @OBIWrapper.checkIsActive             
    def __str__(self) :
        cdef str    to_print
        cdef object line
        to_print = ''
        for line in self :
            to_print = to_print + str(line) + "\n"
        return to_print
 
    
    @OBIWrapper.checkIsActive
    def __repr__(self) :
        cdef str s
        cdef char* sc
        cdef OBIDMS_column_p pointer = self.pointer()
        sc = obi_column_formatted_infos(pointer, False)
        s = bytes2str(sc)
        free(sc)
        return s


    @OBIWrapper.checkIsActive
    def repr_longformat(self) :
        cdef str s
        cdef char* sc
        cdef OBIDMS_column_p pointer = self.pointer()
        sc = obi_column_formatted_infos(pointer, True)
        s = bytes2str(sc)
        free(sc)
        return s
    
         
    def close(self):  # TODO discuss, can't be called bc then bug when closing view that tries to close it in C
    
        cdef OBIDMS_column_p pointer

        if self.active() :
            pointer = self.pointer()
            self._view.unregister(self)
            OBIWrapper.close(self)      
            #if obi_close_column(pointer) < 0 :
            #    raise Exception("Problem closing column %s" % bytes2str(self.name))
 
 
    cdef read_elements_names(self):
        cdef char* elts_names_b
        cdef list  elts_names_list
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        elts_names_b = obi_get_elements_names(self.pointer())
        if elts_names_b == NULL:
            raise Exception("Error reading the elements names of a column")
        elts_names_list = elts_names_b.split(b';')
        free(elts_names_b)
        return elts_names_list
    
    cpdef list keys(self):
        return self._elements_names


    # Column alias property getter and setter
    @property
    def name(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return self._alias
    @name.setter
    def name(self, object new_alias):  # @DuplicatedSignature
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        self._view.rename_column(self._alias, new_alias)

    # elements_names property getter
    @property
    def elements_names(self):
        return self._elements_names
       
    # nb_elements_per_line property getter
    @property
    def nb_elements_per_line(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return self.pointer().header.nb_elements_per_line
 
    # dict_column property getter
    @property
    def dict_column(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return self.pointer().header.dict_column
 
    # data_type property getter
    @property
    def data_type(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return name_data_type(self.data_type_int)
 
    # data_type integer code property getter
    @property
    def data_type_int(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return self.pointer().header.returned_data_type

    # original_name property getter
    @property
    def original_name(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return self.pointer().header.name
 
    # version property getter
    @property
    def version(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return self.pointer().header.version
 
    # lines_used property getter
    @property
    def lines_used(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return self.pointer().header.lines_used
 
    # tuples property getter
    @property
    def tuples(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return self.pointer().header.tuples

    # to_eval property getter
    @property
    def to_eval(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return self.pointer().header.to_eval

    # creation_date property getter
    @property
    def creation_date(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return obi_format_date(self.pointer().header.creation_date)


    # associated_column name property getter and setter
    @property
    def associated_column_name(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return self.pointer().header.associated_column.column_name

    @associated_column_name.setter
    def associated_column_name(self, object new_name):
        strcpy(self.pointer().header.associated_column.column_name, tobytes(new_name))


    # associated_column version property getter and setter
    @property
    def associated_column_version(self):
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        return self.pointer().header.associated_column.version

    @associated_column_version.setter
    def associated_column_version(self, int new_version):
        self.pointer().header.associated_column.version = new_version


    # comments property getter
    @property
    def comments(self):
        return Column_comments(self)
    @comments.setter
    def comments(self, object value):
        Column_comments(self, value)

          
cdef class Column_comments(dict):   # Not thread safe
    def __init__(self, Column column, value=None) :
        if not column.active() :
            raise OBIDeactivatedInstanceError()
        self._column = column
        if value is not None:
            self.update(value)     # TODO test and discuss not overwriting (could use replace bool)
        self._update_from_file()

    def _update_from_file(self):
        cdef bytes comments_json
        cdef str comments_json_str
        cdef OBIDMS_column_p column_p
        cdef Column column
        if not self._column.active() :
            raise OBIDeactivatedInstanceError()
        column = self._column
        column_p = <OBIDMS_column_p>(column.pointer())
        comments_json = <bytes> column_p.header.comments
        comments_json_str = bytes2str(comments_json)
        comments_dict = json.loads(comments_json_str)
        str2bytes_object(comments_dict)
        super(Column_comments, self).update(comments_dict)
        
    def __getitem__(self, object key):
        if not self._column.active() :
            raise OBIDeactivatedInstanceError()
        if type(key) == str:
            key = str2bytes(key)
        self._update_from_file()
        return super(Column_comments, self).__getitem__(key)

    def __setitem__(self, object key, object value):
        cdef OBIDMS_column_p column_p
        cdef Column column

        if not self._column.active() :
            raise OBIDeactivatedInstanceError()

        column = self._column
        column_p = <OBIDMS_column_p>(column.pointer())
        
        # Remove virtually empty values from the object    # TODO discuss
        clean_empty_values_from_object(value)
        
        # If value is virtually empty, don't add it    # TODO discuss
        if value is None or len(value) == 0:
            return
        
        # Convert to bytes
        if type(key) == str:
            key = str2bytes(key)
        value_bytes = str2bytes_object(value)

        # Update dict with comments already written in file
        self._update_from_file()
           
        # Add new element  # TODO don't overwrite?
        super(Column_comments, self).__setitem__(key, value_bytes)
        
        # Convert to str because json library doens't like bytes
        dict_str = {key:item for key,item in self.items()}
        dict_str = bytes2str_object(dict_str)
        
        # Convert to json string
        comments_json = json.dumps(dict_str)        

        # Write new comments
        if obi_column_write_comments(column_p, tobytes(comments_json)) < 0:
            raise Exception("Could not write column comments: %s", comments_json)

    def update(self, value):
        for k,v in value.items():
            self[k] = v

    def __contains__(self, key):
        return super(Column_comments, self).__contains__(tobytes(key))

    def __str__(self):
        return bytes2str(self._column.pointer().header.comments)


######################################################################################################
 

cdef class Column_multi_elts(Column) :
    
    @OBIWrapper.checkIsActive
    def __getitem__(self, index_t line_nb):
        return Column_line(self, line_nb)
    
    cpdef set_line(self, index_t line_nb, object values):
        cdef object value_ref
        cdef object values_iter
        if not self.active() :
            raise OBIDeactivatedInstanceError()
        if values is None or len(values) == 0 or (isinstance(values, Column_line) and values.is_NA()):
            for element_name in self._elements_names :
                self.set_item(line_nb, element_name, None)
        else :
            values_iter = xrange(len(values)) if isinstance(values, list) else values
            for value_ref in values_iter :
                self.set_item(line_nb, value_ref, values[value_ref])


######################################################################################################


cdef class Column_line :
 
    def __init__(self, Column column, index_t line_nb) :
        self._index = line_nb
        self._column = column
         

    def __getitem__(self, object elt_id) :
        return self._column.get_item(self._index, elt_id)


    def __setitem__(self, object elt_id, object value):
        self._column.set_item(self._index, elt_id, value)


    def get(self, object elt_id, object default=None):      # TODO returns default if None???
        if elt_id in self:
            return self._column.get_item(self._index, elt_id)
        else:
            return default
    
    
    def __len__(self):
        return self._column.nb_elements_per_line
         
        
    def __contains__(self, object elt_id):
        if type(elt_id) == int:
            return elt_id < self._column.nb_elements_per_line
        else:
            return (tobytes(elt_id) in self._column._elements_names)
 
 
    def __repr__(self) :
        return str(self._column.get_line(self._index))
 
 
    def is_NA(self):
        return self._column.get_line(self._index) is None
    
    
    def __iter__(self) :
        cdef list  elements_names
        cdef bytes element_name
        for element_name in self._column.get_line(self._index) :
            yield element_name
    
    
    cpdef list keys(self):    
    # TODO just returning all element names of the column would be faster but 
    # potentially lots of keys associated with None values on that given line.
    # get_line() methods don't return None values
        return [element_name for element_name in self._column.get_line(self._index)]
    

    cpdef list items(self):
        return [(k, self[k]) for k in self.keys()]
    
    
    cpdef dict dict(self):
        cdef dict d
        d = {}
        for k,v in self.items():
            d[k] = v
        return d
    
    
    def __str__(self):
        cdef dict d
        cdef str  s
        if self.is_NA():
            return ""
        d = {}
        for k,v in self.items():
            if type(v) == bytes:
                value = bytes2str(v)
            else:
                value = v
            d[bytes2str(k)] = value
        s = str(d)
        return s


    cpdef bytes bytes(self):
        cdef str   s
        cdef bytes b
        s = self.__str__()
        b = str2bytes(s)
        return b
        
        
    # column property getter
    @property
    def column(self):
        return self._column

    # index property getter
    @property
    def index(self):
        return self._index
        
    
    cpdef update(self, object data):
        if isinstance(data, dict):
            data=data.items()
        for key,value in data:
            if key in self:
                self[key]=value

 
######################################################################################################


cdef register_column_class(obitype_t obitype,
                           bint multi_elts,
                           bint tuples,
                           type obiclass,
                           type python_type):
    '''
    Each sub class of `OBIDMS_column` needs to be registered after its declaration
    to declare its relationship with an `OBIType_t`
    '''
    global __OBIDMS_COLUMN_CLASS__
    
    assert issubclass(obiclass, Column)
    
    __OBIDMS_COLUMN_CLASS__[(obitype, multi_elts, tuples)] = (obiclass, python_type)


cdef register_all_column_classes() :
    
    x = list(pkgutil.walk_packages(typed_column.__path__, prefix="obitools3.dms.column.typed_column."))
    all_modules = [importlib.import_module(a[1]) for a in x]
    for mod in all_modules :
        getattr(mod, 'register_class')()


register_all_column_classes()



    


#cython: language_level=3

from ._dynamic cimport *

cdef class NWS(DynamicProgramming):
    cdef double _match
    cdef double _mismatch
    
    cdef double matchScore(self,int h, int v)
    cdef double doAlignment(self) except? 0

             



#cython: language_level=3

'''
Created on 6 Nov. 2009

@author: coissac
'''


cdef class RightDirectAssemble(NWS):
            
    def __init__(self,match=4,mismatch=-6,opengap=-8,extgap=-2):
        NWS.__init__(self,match,mismatch,opengap,extgap)
        self.xsmax=0
        self.xmax=0
                
    cdef double doAlignment(self) except? 0:
        cdef int i  # vertical index
        cdef int j  # horizontal index
        cdef int idx
        cdef int jump
        cdef int delta
        cdef double score
        cdef double scoremax
        cdef int    path

        
        if self.needToCompute:
            self.allocate()
            self.reset()
            self.xsmax=0
            self.xmax=0

            for j in range(1,self.hSeq.length+1):
                idx = self.index(j,0)
                self.matrix.matrix[idx].score = self._opengap + (self._extgap * (j-1))
                self.matrix.matrix[idx].path  = j
                                
            for i in range(1,self.vSeq.length+1):
                idx = self.index(0,i)
                self.matrix.matrix[idx].score = 0
                self.matrix.matrix[idx].path  = -i
                
            for i in range(1,self.vSeq.length+1):
                for j in range(1,self.hSeq.length+1):
                                        
                    # 1 - came from diagonal
                    idx = self.index(j-1,i-1)
                    #print("computing cell : %d,%d --> %d/%d" % (j,i,self.index(j,i),self.matrix.msize),)
                    scoremax = self.matrix.matrix[idx].score + \
                               self.matchScore(j,i)
                    path = 0

                    #print("so=%f sd=%f sm=%f" % (self.matrix.matrix[idx].score,self.matchScore(j,i),scoremax),)

                    # 2 - open horizontal gap
                    idx = self.index(j-1,i)
                    score = self.matrix.matrix[idx].score+ \
                            self._opengap
                    if score > scoremax : 
                        scoremax = score
                        path = +1
                    
                    # 3 - open vertical gap
                    idx = self.index(j,i-1)
                    score = self.matrix.matrix[idx].score + \
                            self._opengap
                    if score > scoremax : 
                        scoremax = score
                        path = -1
                        
                    # 4 - extend horizontal gap
                    jump = self.matrix.bestHJump[i]
                    if jump >= 0:
                        idx = self.index(jump,i)
                        delta = j-jump
                        score = self.matrix.matrix[idx].score + \
                                self._extgap * delta
                        if score > scoremax :
                            scoremax = score
                            path = delta+1 
                            
                    # 5 - extend vertical gap
                    jump = self.matrix.bestVJump[j]
                    if jump >= 0:
                        idx = self.index(j,jump)
                        delta = i-jump
                        score = self.matrix.matrix[idx].score + \
                                self._extgap * delta
                        if score > scoremax :
                            scoremax = score
                            path = -delta-1 
    
                    idx = self.index(j,i)
                    self.matrix.matrix[idx].score = scoremax
                    self.matrix.matrix[idx].path  = path 

                    if path == -1:
                        self.matrix.bestVJump[j]=i
                    elif path == +1 :
                        self.matrix.bestHJump[i]=j

                    if i==self.vSeq.length and scoremax > self.xsmax:
                        self.xsmax=scoremax
                        self.xmax=j
                                                         
        self.sequenceChanged=False
        self.scoreChanged=False
                
        return self.xsmax
                   
    cdef void backtrack(self):
        cdef list path=[]
        cdef int i
        cdef int j 
        cdef int p
        
        self.doAlignment()
        j=self.xmax
        i=self.vSeq.length
        self.path=allocatePath(i,j+1,self.path)
        
        if self.xmax<self.hSeq.length:
            self.path.path[self.path.length]=self.hSeq.length-self.xmax
            self.path.length+=1
            
        while (i or j):
            p=self.matrix.matrix[self.index(j,i)].path
            self.path.path[self.path.length]=p
            self.path.length+=1
            #path.append(p)
            if p==0:
                i-=1
                j-=1
            elif p < 0:
                i+=p
            else:
                j-=p
                
        #path.reverse()
        self.path.hStart=0
        self.path.vStart=0
        #reversePath(self.path)
        #return 0,0,path
                           
        
cdef class RightReverseAssemble(RightDirectAssemble):    

    property seqB:
            def __get__(self):
                return self.verticalSeq.wrapped
            
            def __set__(self, seq):
                self.sequenceChanged=True
                self.verticalSeq=seq.reverse_complement
                self.vSeq=allocateSequence(self.verticalSeq,self.vSeq)



#cython: language_level=3


cdef extern from "build_reference_db.h" nogil:

    int build_reference_db(const char* dms_name,
                           const char* refs_view_name,
                           const char* taxonomy_name,
                           const char* o_view_name,
                           const char* o_view_comments,
                           double threshold)




#cython: language_level=3

from obitools3.dms.capi.obiview cimport NUC_SEQUENCE_COLUMN, QUALITY_COLUMN

from obitools3.dms.capi.kmer_similarity cimport kmer_similarity, Obi_ali_p, obi_free_shifted_ali
from obitools3.dms.view.typed_view.view_NUC_SEQS cimport View_NUC_SEQS
from obitools3.dms.column.column cimport Column

from libc.stdlib  cimport free


cdef class Ali_shifted:
            
    cdef inline Obi_ali_p pointer(self) :
        return <Obi_ali_p>(self._pointer)        
    
    @staticmethod
    cdef Ali_shifted new_ali(Obi_ali_p ali_p) :   # not __init__ method because need to pass pointer
        ali = Ali_shifted()
        ali._pointer = ali_p
        return ali

    @property
    def score(self):
        return self.pointer().score

    @property
    def consensus_len(self):
        return self.pointer().consensus_length

    @property
    def overlap_len(self):
        return self.pointer().overlap_length
    
    @property
    def consensus_seq(self):
        return self.pointer().consensus_seq

    @property
    def shift(self):
        return self.pointer().shift
    
    @property
    def consensus_qual(self):   # must return list because uint8_t* are forced into bytes by cython which won't keep beyond the first 0 value
        qual_list = []
        for i in range(self.consensus_len):
            qual_list.append((self.pointer().consensus_qual)[i])
        return qual_list

    @property
    def direction(self):
        return self.pointer().direction

    cpdef free(self):  # TODO allocated memory could be kept
        obi_free_shifted_ali(self.pointer())


cdef class Kmer_similarity:
    def __init__(self, View_NUC_SEQS view1, Column column1=None, Column qual_column1=None, \
                 View_NUC_SEQS view2=None, Column column2=None, Column qual_column2=None, \
                 uint8_t kmer_size=3, bint build_consensus=True, Column reversed_column=None) :
        cdef Column default_seq_col
        cdef Column default_qual_col
        if kmer_size < 1 or kmer_size > 3:
            raise Exception("Kmer size to compute kmer similarity must be >=1 or <=4")
        self.kmer_pos_array_p = NULL
        self.shift_array_p = NULL
        self.shift_count_array_p = NULL
        self.kmer_pos_array_height_a[0] = 0
        self.shift_array_height_a[0] = 0 
        self.shift_count_array_height_a[0] = 0
        self.kmer_size = kmer_size
        self.build_consensus = build_consensus
        self.view1_p = view1.pointer()
        if column1 is not None:
            self.column1_p = column1.pointer()
        else:
            if type(view1) != View_NUC_SEQS:
                raise Exception("Kmer similarity with no sequence column given must be given a NUC_SEQS view")
            default_seq_col = view1[NUC_SEQUENCE_COLUMN]
            self.column1_p = default_seq_col.pointer()
        if view2 is not None:
            self.view2_p = view2.pointer()
        else:
            view2 = view1
            self.view2_p = self.view1_p
        if column2 is not None:
            self.column2_p = column2.pointer()
        else:
            if type(view2) != View_NUC_SEQS:
                raise Exception("Kmer similarity with no sequence column given must be given a NUC_SEQS view")
            default_seq_col = view2[NUC_SEQUENCE_COLUMN]
            self.column2_p = default_seq_col.pointer()
        if qual_column1 is not None:
            self.qual_col1_p = qual_column1.pointer()
        else:
            if type(view1) != View_NUC_SEQS:
                raise Exception("Kmer similarity with no quality column given must be given a NUC_SEQS view")
            default_qual_col = view1[QUALITY_COLUMN]
            self.qual_col1_p = default_qual_col.pointer()
        if qual_column2 is not None:
            self.qual_col2_p = qual_column2.pointer()
        else:
            if type(view2) != View_NUC_SEQS:
                raise Exception("Kmer similarity with no quality column given must be given a NUC_SEQS view")
            default_qual_col = view2[QUALITY_COLUMN]
            self.qual_col2_p = default_qual_col.pointer()
        if reversed_column is None:
            self.reversed_col_p = NULL
        else:
            self.reversed_col_p = reversed_column.pointer()

    
    def __call__(self, object seq1, object seq2):
        cdef Obi_ali_p ali_p
        cdef Ali_shifted ali
        ali_p = kmer_similarity(self.view1_p, self.column1_p, seq1.index, 0, \
                               self.view2_p, self.column2_p, seq2.index, 0, \
                               self.qual_col1_p, self.qual_col2_p, \
                               self.reversed_col_p, \
                               self.kmer_size, \
                               &(self.kmer_pos_array_p), self.kmer_pos_array_height_a, \
                               &(self.shift_array_p), self.shift_array_height_a, \
                               &(self.shift_count_array_p), self.shift_count_array_height_a,
                               self.build_consensus)
        
        ali = Ali_shifted.new_ali(ali_p)
        return ali
    
    cpdef free(self):
        if self.kmer_pos_array_p is not NULL:
            free(self.kmer_pos_array_p)
            self.kmer_pos_array_height_a[0] = 0
        if (self.shift_array_p is not NULL) :
            free(self.shift_array_p)
            self.shift_array_height_a[0] = 0
        if (self.shift_count_array_p is not NULL) :
            free(self.shift_count_array_p)
            self.shift_count_array_height_a[0] = 0
        




#cython: language_level=3

from ._nws cimport *  

cdef class DirectAssemble(NWS):
    cdef double ysmax
    cdef int    ymax

    cdef double doAlignment(self) except? 0
    
cdef class ReverseAssemble(DirectAssemble):    
    pass


#cython: language_level=3

'''
Created on 30 mars 2016

@author: coissac
'''

import types

from obitools3.dms.obiseq cimport Nuc_Seq


# def fastaIterator(lineiterator, 
#                   int skip=0,
#                   only=None,
#                   firstline=None,
#                   int buffersize=100000000
#                  ):
#     cdef str        ident
#     cdef str        definition
#     cdef dict       tags
#     cdef list       s
#     cdef bytes      sequence
#     cdef int        skipped, ionly, read
#     
#     if only is None:
#         ionly=-1
#     else:
#         ionly=int(only)
# 
#     if isinstance(lineiterator, (str, bytes)):
#         lineiterator=uopen(lineiterator)        
#     if isinstance(lineiterator, LineBuffer):
#         iterator = iter(lineiterator)
#     else:
#         if hasattr(lineiterator, "readlines"):
#             iterator = iter(LineBuffer(lineiterator, buffersize))
#         elif hasattr(lineiterator, '__next__'):
#             iterator = lineiterator
#         else:
#             raise Exception("Invalid line iterator")
#     
#     skipped = 0
#     i = iterator
#     
#     if firstline is None:
#         line = next(i)
#     else:
#         line = firstline
#         
#     while True:
#         
#         if ionly >= 0 and read >= ionly:
#             break
#                 
#         while skipped < skip :
#             line = next(i)
#             try:
#                 while line[0]!='>':
#                     line = next(i)
#             except StopIteration:
#                 pass
#             skipped += 1
# 
#         ident,tags,definition = parseHeader(line)
#         s = []
#         line = next(i)
#             
#         try:
#             while line[0]!='>':
#                 s.append(str2bytes(line)[0:-1])
#                 line = next(i)
#         
#         except StopIteration:
#             pass
#         
#         sequence  = b"".join(s)
# 
#         yield { "id"         : ident,
#                 "definition" : definition,
#                 "sequence"   : sequence,
#                 "quality"    : None,
#                 "offset"     : None,
#                 "tags"       : tags,
#                 "annotation" : {}
#               }
#         
#         read+=1

    
def fastaNucIterator(lineiterator, 
                     int skip=0,
                     only=None,
                     firstline=None,
                     int buffersize=100000000,
                     bytes nastring=b"NA"
                    ):
    
    cdef bytes      ident
    cdef bytes      definition
    cdef dict       tags
    cdef list       s
    cdef bytes      sequence
    cdef int        skipped, ionly, read
    cdef Nuc_Seq    seq
    cdef bint       stop
    
    if only is None:
        ionly = -1
    else:
        ionly = int(only)

    if isinstance(lineiterator, (str, bytes)):
        lineiterator=uopen(lineiterator)        
    if isinstance(lineiterator, LineBuffer):
        iterator = iter(lineiterator)
    else:
        if hasattr(lineiterator, "readlines"):
            iterator = iter(LineBuffer(lineiterator, buffersize))
        elif hasattr(lineiterator, '__next__'):
            iterator = lineiterator
        else:
            raise Exception("Invalid line iterator")

    skipped = 0
    read = 0
    
    if firstline is None:
        line = next(iterator)
    else:
        line = firstline       
    
    stop=False
    while not stop:
                
        if ionly >= 0 and read >= ionly:
            break
                
        while skipped < skip :
            line = next(iterator)
            try:
                while line[:1]!=b'>':
                    line = next(iterator)
            except StopIteration:
                pass
            skipped += 1

        ident,tags,definition = parseHeader(line, nastring=nastring)
        s = []
        line = next(iterator)
    
        try:
            while line[:1]!=b'>':
                s.append(line[0:-1])
                line = next(iterator)
        except StopIteration:
            stop=True
        
        sequence  = b"".join(s)        
        
        seq = Nuc_Seq(ident,
                      sequence,
                      definition=definition,
                      quality=None,
                      offset=-1,
                      tags=tags)
            
        yield seq
        
        read+=1
        
        


#cython: language_level=3

from ._nws cimport *  

cdef class RightDirectAssemble(NWS):
    cdef double xsmax
    cdef int    xmax

    cdef double doAlignment(self) except? 0
    
cdef class RightReverseAssemble(RightDirectAssemble):    
    pass


#cython: language_level=3

from .view.view cimport Line


cdef class Seq(dict) :

    cdef int _index
    cpdef object clone(self)
    cpdef str get_str(self)
    cpdef get_symbol_at(self, int pos)
    cpdef get_slice(self, slice slice_to_get)
    
    
cdef class Nuc_Seq(Seq) :
    
    cdef  Nuc_Seq _reverse_complement
    cdef  object  _quality_array
    cdef  bint    _is_revcomp
    cpdef set_quality(self, object new_quality, int offset=*)  
    cpdef object build_quality_array(self, list quality)
    cpdef bytes  build_reverse_complement(self)
    

cdef class Seq_Stored(Line) :

    cpdef get_symbol_at(self, int pos)
    cpdef get_slice(self, slice slice_to_get)


cdef class Nuc_Seq_Stored(Seq_Stored) :
    
    cdef  Nuc_Seq _reverse_complement
    cdef  object  _quality_array
    cdef  bytes   _seq
    
    cpdef set(self, object id, object seq, object definition=*, object quality=*, int offset=*, object tags=*)
    cpdef set_quality_int(self, list new_qual)
    cpdef set_quality_char(self, object new_qual, int offset=*)
    cpdef object build_quality_array(self, list quality)
    cpdef bytes build_reverse_complement(self)
    cpdef str get_str(self)
    cpdef repr_bytes(self)
    


#cython: language_level=3

from obitools3.dms.capi.obierrno cimport obi_errno

from ..column cimport register_column_class

from ...view.view cimport View

from obitools3.utils cimport tobytes, \
                             obi_errno_to_exception

from ...capi.obiview cimport obi_get_bool_with_elt_name_and_col_p_in_view, \
                             obi_get_bool_with_elt_idx_and_col_p_in_view, \
                             obi_set_bool_with_elt_name_and_col_p_in_view, \
                             obi_set_bool_with_elt_idx_and_col_p_in_view, \
                             obi_get_array_with_col_p_in_view, \
                             obi_set_array_with_col_p_in_view, \
                             Obiview_p

from ...capi.obidmscolumn cimport OBIDMS_column_p

from ...capi.obitypes cimport OBI_BOOL, \
                              OBIBool_NA, \
                              OBITuple_NA, \
                              obibool_t

from cpython.bool cimport PyBool_FromLong

from libc.stdint cimport int32_t

from libc.stdlib  cimport malloc, free


cdef class Column_bool(Column):

    @staticmethod
    def new(View    view,
            object  column_name,
            index_t nb_elements_per_line=1,
            object  elements_names=None,
            bint    dict_column=False,
            bint    tuples=False,
            object  comments={}):
        return Column.new_column(view, column_name, OBI_BOOL,
                          nb_elements_per_line=nb_elements_per_line, 
                          elements_names=elements_names,
                          dict_column=dict_column, 
                          tuples=tuples,
                          comments=comments)
                
    cpdef object get_line(self, index_t line_nb):
        global obi_errno
        cdef   obibool_t value
        cdef   object    result
        value = obi_get_bool_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0)
        obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem getting a value from a column")
        if value == OBIBool_NA :
            result = None
        else :
            result = PyBool_FromLong(value)
        return result
    
    cpdef set_line(self, index_t line_nb, object value):
        global obi_errno
        if value is None :
            value = OBIBool_NA
        if obi_set_bool_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0, <obibool_t> value) < 0 :
            obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem setting a value in a column")
 
 
cdef class Column_multi_elts_bool(Column_multi_elts):
 
    cpdef object get_item(self, index_t line_nb, object elt_id) :
        global obi_errno
        cdef   obibool_t value
        cdef   object    result
        cdef   bytes     elt_name
        if type(elt_id) == int :
            value = obi_get_bool_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_id)
        else :
            elt_name = tobytes(elt_id)
            value = obi_get_bool_with_elt_name_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_name)
        obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem getting a value from a column")
        if value == OBIBool_NA :
            result = None
        else :
            result = PyBool_FromLong(value)
        return result
    
    cpdef object get_line(self, index_t line_nb) :
        global obi_errno
        cdef   obibool_t       value
        cdef   bint            value_in_result
        cdef   dict            result
        cdef   index_t         i
        cdef   bint            all_NA
        cdef   list            elements_names
        cdef   Obiview_p       view_p
        cdef   OBIDMS_column_p column_p
        result = {}
        all_NA = True
        view_p = self._view.pointer()
        column_p = self.pointer()
        elements_names = self._elements_names
        for i in range(self.nb_elements_per_line) :
            value = obi_get_bool_with_elt_idx_and_col_p_in_view(view_p, column_p, line_nb, i)
            obi_errno_to_exception(line_nb=line_nb, elt_id=i, error_message="Problem getting a value from a column")
            if value != OBIBool_NA :
                value_in_result = PyBool_FromLong(value)   
                result[elements_names[i]] = value_in_result
                if all_NA :
                    all_NA = False
        if all_NA :
            result = None
        return result
    
    cpdef set_item(self, index_t line_nb, object elt_id, object value) :
        global obi_errno
        cdef   bytes elt_name
        if value is None :
            value = OBIBool_NA
        if type(elt_id) == int :
            if obi_set_bool_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_id, <obibool_t> value) < 0 :
                obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem setting a value in a column")
        else :
            elt_name = tobytes(elt_id)
            if obi_set_bool_with_elt_name_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_name, <obibool_t> value) < 0 :
                obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem setting a value in a column")


cdef class Column_tuples_bool(Column):
    
    cpdef object get_line(self, index_t line_nb) :
        global obi_errno
        cdef   obibool_t       value
        cdef   bint            value_in_result
        cdef   object          result
        cdef   int32_t         i
        cdef   obibool_t*      array
        cdef   int32_t         value_length
        
        result = []
        
        array = <obibool_t*>obi_get_array_with_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, &value_length)
        obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem getting a value from a column")

        if array == OBITuple_NA :
            return None

        for i in range(value_length) :
            value = array[i]
            value_in_result = PyBool_FromLong(value)
            result.append(value_in_result)
        
        return tuple(result)


    cpdef set_line(self, index_t line_nb, object value) :
        global             obi_errno
        cdef   obibool_t*  array
        cdef   int32_t     value_length
        cdef   int32_t     i, j
        cdef   object      e
            
        value_length = 0
        if value is not None:
            for e in value:
                if e is not None:
                    value_length+=1
        if value is None or value_length == 0 :
            array = <obibool_t*>OBITuple_NA
        else:
            array = <obibool_t*>malloc(value_length * sizeof(obibool_t))
            if array == NULL:
                raise Exception("Problem allocating memory for an array to store a tuple")
                #raise RollbackException("Problem allocating memory for an array to store a tuple", self._view) # TODO can't import
            j=0
            for i in range(len(value)) :
                if value[i] is not None:
                    array[j] = <obibool_t>(value[i])
                    j+=1

        if obi_set_array_with_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, <obibool_t*> array, sizeof(obibool_t)*8, value_length) < 0 :
            obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem setting a value in a column")

        if array != <obibool_t*>OBITuple_NA:
            free(array)
        

def register_class() :
    register_column_class(OBI_BOOL, False, False, Column_bool, bool)
    register_column_class(OBI_BOOL, True, False, Column_multi_elts_bool, bool)
    register_column_class(OBI_BOOL, False, True, Column_tuples_bool, bool)




#cython: language_level=3


cdef extern from "obi_ecotag.h" nogil:

    int obi_ecotag(const char* dms_name,
                   const char* query_view_name,
                   const char* ref_dms_name,
                   const char* ref_view_name,
                   const char* taxo_dms_name,
                   const char* taxonomy_name,
                   const char* output_view_name,
                   const char* output_view_comments,
                   double ecotag_threshold,
                   double bubble_threshold)



#cython: language_level=3

cimport cython
from obitools3.dms.capi.obiview cimport NUC_SEQUENCE_COLUMN
from obitools3.utils cimport bytes2str


cdef class FastaFormat:
	
	def __init__(self, list tags=[], bint printNAKeys=False, bytes NAString=b"NA", bint NAIntTo0=False):
		self.headerFormatter = HeaderFormat("fasta",
										    tags=tags,
										    printNAKeys=printNAKeys,
										    NAString=NAString,
										    NAIntTo0=NAIntTo0)
		
	@cython.boundscheck(False)	
	def __call__(self, object data):
		cdef bytes brawseq      = data[NUC_SEQUENCE_COLUMN]
		cdef size_t lseq        = len(brawseq)
		cdef size_t k           = 0
		cdef list   lines       = []
		
		for k in range(0,lseq,60):
			lines.append(brawseq[k:(k+60)])
			
		brawseq = b'\n'.join(lines)
		
		return self.headerFormatter(data) + b"\n" + brawseq
					



#cython: language_level=3

from obitools3.dms.capi.obierrno cimport obi_errno

from ..view.view cimport View

from obitools3.utils cimport tobytes, \
                             obi_errno_to_exception

from ..capi.obiview cimport obi_get_index_with_elt_name_and_col_p_in_view, \
                             obi_get_index_with_elt_idx_and_col_p_in_view, \
                             obi_set_index_with_elt_name_and_col_p_in_view, \
                             obi_set_index_with_elt_idx_and_col_p_in_view, \
                             Obiview_p

from ..capi.obidmscolumn cimport OBIDMS_column_p

from ..capi.obitypes     cimport OBI_IDX, OBIIdx_NA, index_t

from cpython.long cimport PyLong_FromLongLong


# TODO overwrite other functions from Column and Column_multi_elts


cdef class Column_idx(Column):

    cpdef object get_line_idx(self, index_t line_nb):
        global obi_errno
        cdef   index_t value
        cdef   object  result 
        value = obi_get_index_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0)
        obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem getting a value from a column")
        if value == OBIIdx_NA :
            result = None
        else :
            result = PyLong_FromLongLong(value)
        return result

    cpdef set_line_idx(self, index_t line_nb, object value):
        global obi_errno
        if value is None :
            value = OBIIdx_NA
        if obi_set_index_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0, <index_t> value) < 0:
            obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem setting a value in a column")
 
 
cdef class Column_multi_elts_idx(Column_multi_elts):
 
    cpdef object get_item_idx(self, index_t line_nb, object elt_id) :
        global obi_errno
        cdef   int       value
        cdef   object    result
        cdef   bytes     elt_name
        if type(elt_id) == int :
            value = obi_get_index_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_id)
        else :
            elt_name = tobytes(elt_id)
            value = obi_get_index_with_elt_name_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_name)
        obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem getting a value from a column")
        if value == OBIIdx_NA :
            result = None
        else :
            result = PyLong_FromLongLong(value)
        return result
    
    cpdef object get_line_idx(self, index_t line_nb) :
        global obi_errno
        cdef   index_t         value
        cdef   int             value_in_result
        cdef   dict            result
        cdef   index_t         i
        cdef   bint            all_NA
        cdef   list            elements_names
        cdef   Obiview_p       view_p
        cdef   OBIDMS_column_p column_p
        result = {}
        all_NA = True
        view_p = self._view.pointer()
        column_p = self.pointer()
        elements_names = self._elements_names
        for i in range(self.nb_elements_per_line) :
            value = obi_get_index_with_elt_idx_and_col_p_in_view(view_p, column_p, line_nb, i)
            obi_errno_to_exception(line_nb=line_nb, elt_id=i, error_message="Problem getting a value from a column")
            if value != OBIIdx_NA :
                value_in_result = PyLong_FromLongLong(value)
                result[elements_names[i]] = value_in_result
                if all_NA :
                    all_NA = False
        if all_NA :
            result = None
        return result
    
    cpdef set_item_idx(self, index_t line_nb, object elt_id, object value) :
        global obi_errno
        cdef   bytes elt_name
        if value is None :
            value = OBIIdx_NA
        if type(elt_id) == int :
            if obi_set_index_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_id, <index_t> value) < 0 :
                obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem setting a value in a column")
        else :
            elt_name = tobytes(elt_id)
            if obi_set_index_with_elt_name_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_name, <index_t> value) < 0 :
                obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem setting a value in a column")

    cpdef set_line_idx(self, index_t line_nb, object values):
        if values is None :
            for element_name in self._elements_names :
                self.set_item_idx(line_nb, element_name, None)
        else :
            for element_name in values :
                self.set_item_idx(line_nb, element_name, values[element_name])




#cython: language_level=3

'''
Created on 25 mars 2016

@author: coissac
'''

from urllib.request import urlopen
from obitools3.utils cimport tostr


cpdef CompressedFile uopen(object name, mode='rb'):
    cdef CompressedFile c
    
    try:
        f = urlopen(tostr(name))
    except:
        f = open(tostr(name),mode) 
        
    c = CompressedFile(f)
    
    return c
    


#cython: language_level=3

from .obitypes           cimport const_char_p, \
                                 OBIType_t, \
                                 obiversion_t, \
                                 obiint_t, \
                                 obibool_t, \
                                 obichar_t, \
                                 obifloat_t, \
                                 index_t, \
                                 time_t
from ..capi.obidms       cimport OBIDMS_p
from ..capi.obidmscolumn cimport OBIDMS_column_p, \
                                 Column_reference_t, \
                                 Column_reference_p

from libc.stdint cimport uint8_t, int32_t


cdef extern from "obiview.h" nogil:

    extern const_char_p VIEW_TYPE_NUC_SEQS
    extern const_char_p NUC_SEQUENCE_COLUMN
    extern const_char_p ID_COLUMN
    extern const_char_p DEFINITION_COLUMN
    extern const_char_p QUALITY_COLUMN
    extern const_char_p REVERSE_QUALITY_COLUMN
    extern const_char_p REVERSE_SEQUENCE_COLUMN
    extern const_char_p COUNT_COLUMN
    extern const_char_p SCIENTIFIC_NAME_COLUMN
    extern const_char_p TAXID_COLUMN
    extern const_char_p MERGED_TAXID_COLUMN
    extern const_char_p MERGED_PREFIX
    extern const_char_p TAXID_DIST_COLUMN
    extern const_char_p MERGED_COLUMN

    struct Alias_column_pair_t :
        Column_reference_t    column_refs
        const_char_p          alias
    
    ctypedef Alias_column_pair_t* Alias_column_pair_p


    struct Obiview_infos_t :
        time_t              creation_date
        const_char_p        name
        const_char_p        created_from
        const_char_p        view_type
        bint                all_lines
        Column_reference_t  line_selection
        index_t             line_count
        int                 column_count
        Alias_column_pair_p column_references
        const_char_p        comments

    ctypedef Obiview_infos_t* Obiview_infos_p


    struct Obiview_t :
        Obiview_infos_p     infos
        OBIDMS_p            dms
        bint                read_only
        OBIDMS_column_p     line_selection
        OBIDMS_column_p     columns
        int                 nb_predicates
        # TODO declarations for column dictionary and predicate function array?
        
    ctypedef Obiview_t* Obiview_p


    Obiview_p obi_new_view_nuc_seqs(OBIDMS_p dms, const_char_p view_name, Obiview_p view_to_clone, index_t* line_selection, const_char_p comments, bint quality_column, bint create_default_columns)

    Obiview_p obi_new_view(OBIDMS_p dms, const_char_p view_name, Obiview_p view_to_clone, index_t* line_selection, const_char_p comments)
    
    Obiview_p obi_new_view_cloned_from_name(OBIDMS_p dms, const_char_p view_name, const_char_p view_to_clone_name, index_t* line_selection, const_char_p comments)

    Obiview_p obi_new_view_nuc_seqs_cloned_from_name(OBIDMS_p dms, const_char_p view_name, const_char_p view_to_clone_name, index_t* line_selection, const_char_p comments)
    
    Obiview_p obi_clone_view(OBIDMS_p dms, Obiview_p view_to_clone, const char* view_name, index_t* line_selection, const char* comments)

    Obiview_p obi_clone_view_from_name(OBIDMS_p dms, const char* view_to_clone_name, const char* view_name, index_t* line_selection, const char* comments)
    
    Obiview_infos_p obi_view_map_file(OBIDMS_p dms, const char* view_name, bint finished)

    int obi_view_unmap_file(OBIDMS_p dms, Obiview_infos_p view_infos)

    Obiview_p obi_open_view(OBIDMS_p dms, const_char_p view_name)
    
    int obi_view_add_column(Obiview_p view,
                            char* column_name,
                            obiversion_t version_number,
                            const_char_p alias,
                            OBIType_t data_type,
                            index_t nb_lines,
                            index_t nb_elements_per_line,
                            char* elements_names,
                            bint  elt_names_formatted,
                            bint dict_column,
                            bint tuples,
                            bint to_eval,
                            const_char_p indexer_name,
                            const_char_p associated_column_name,
                            obiversion_t associated_column_version,
                            const_char_p comments,
                            bint create)

    int obi_view_delete_column(Obiview_p view, const_char_p column_name, bint delete_file)

    OBIDMS_column_p obi_view_get_column(Obiview_p view, const_char_p column_name)

    OBIDMS_column_p* obi_view_get_pointer_on_column_in_view(Obiview_p view, const_char_p column_name)

    int obi_view_create_column_alias(Obiview_p view, const_char_p current_name, const_char_p alias)    

    char* obi_view_formatted_infos(Obiview_p view, bint detailed)

    char* obi_view_formatted_infos_one_line(Obiview_p view)

    int obi_view_write_comments(Obiview_p view, const_char_p comments)

    int obi_view_add_comment(Obiview_p view, const_char_p key, const_char_p value)

    int obi_save_and_close_view(Obiview_p view)

    int obi_clean_unfinished_views(OBIDMS_p dms)

    int obi_rollback_view(Obiview_p view)

    int obi_delete_view(OBIDMS_p dms, const char* view_name)


    # OBI_INT
    int obi_set_int_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                    OBIDMS_column_p column_p, 
                                                    index_t line_nb, 
                                                    const_char_p element_name, 
                                                    obiint_t value)

    int obi_set_int_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                   OBIDMS_column_p column_p, 
                                                   index_t line_nb, 
                                                   index_t element_idx, 
                                                   obiint_t value)
    
    obiint_t obi_get_int_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                         OBIDMS_column_p column_p, 
                                                         index_t line_nb, 
                                                         const_char_p element_name)
    
    obiint_t obi_get_int_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                        OBIDMS_column_p column_p, 
                                                        index_t line_nb, 
                                                        index_t element_idx)


    # OBI_BOOL
    int obi_set_bool_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                     OBIDMS_column_p column_p, 
                                                     index_t line_nb, 
                                                     const_char_p element_name, 
                                                     obibool_t value)

    int obi_set_bool_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                    OBIDMS_column_p column_p, 
                                                    index_t line_nb, 
                                                    index_t element_idx, 
                                                    obibool_t value)
    
    obibool_t obi_get_bool_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                           OBIDMS_column_p column_p, 
                                                           index_t line_nb, 
                                                           const_char_p element_name)
    
    obibool_t obi_get_bool_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                          OBIDMS_column_p column_p, 
                                                          index_t line_nb, 
                                                          index_t element_idx)


    # OBI_CHAR
    int obi_set_char_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                     OBIDMS_column_p column_p, 
                                                     index_t line_nb, 
                                                     const_char_p element_name, 
                                                     obichar_t value)

    int obi_set_char_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                    OBIDMS_column_p column_p, 
                                                    index_t line_nb, 
                                                    index_t element_idx, 
                                                    obichar_t value)
    
    obichar_t obi_get_char_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                           OBIDMS_column_p column_p, 
                                                           index_t line_nb, 
                                                           const_char_p element_name)
    
    obichar_t obi_get_char_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                          OBIDMS_column_p column_p, 
                                                          index_t line_nb, 
                                                          index_t element_idx)


    # OBI_FLOAT
    int obi_set_float_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                      OBIDMS_column_p column_p, 
                                                      index_t line_nb, 
                                                      const_char_p element_name, 
                                                      obifloat_t value)

    int obi_set_float_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                     OBIDMS_column_p column_p, 
                                                     index_t line_nb, 
                                                     index_t element_idx, 
                                                     obifloat_t value)
    
    obifloat_t obi_get_float_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                             OBIDMS_column_p column_p, 
                                                             index_t line_nb, 
                                                             const_char_p element_name)
    
    obifloat_t obi_get_float_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                            OBIDMS_column_p column_p, 
                                                            index_t line_nb, 
                                                            index_t element_idx)


    # OBI_QUAL
    int obi_set_qual_char_with_elt_idx_and_col_p_in_view(Obiview_p view, 
                                                         OBIDMS_column_p column_p, 
                                                         index_t line_nb, 
                                                         index_t element_idx, 
                                                         const char* value,
                                                         int offset)

    int obi_set_qual_int_with_elt_idx_and_col_p_in_view(Obiview_p view, 
                                                        OBIDMS_column_p column_p, 
                                                        index_t line_nb, 
                                                        index_t element_idx, 
                                                        const uint8_t* value, 
                                                        int value_length)

    char* obi_get_qual_char_with_elt_idx_and_col_p_in_view(Obiview_p view, 
                                                           OBIDMS_column_p column_p, 
                                                           index_t line_nb, 
                                                           index_t element_idx,
                                                           int offset)

    const uint8_t* obi_get_qual_int_with_elt_idx_and_col_p_in_view(Obiview_p view, 
                                                                   OBIDMS_column_p column_p, 
                                                                   index_t line_nb, 
                                                                   index_t element_idx, 
                                                                   int* value_length)

    int obi_set_qual_char_with_elt_name_and_col_p_in_view(Obiview_p view, 
                                                          OBIDMS_column_p column_p, 
                                                          index_t line_nb, 
                                                          const char* element_name, 
                                                          const char* value,
                                                          int offset)

    int obi_set_qual_int_with_elt_name_and_col_p_in_view(Obiview_p view, 
                                                         OBIDMS_column_p column_p, 
                                                         index_t line_nb, 
                                                         const char* element_name, 
                                                         const uint8_t* value, 
                                                         int value_length)

    char* obi_get_qual_char_with_elt_name_and_col_p_in_view(Obiview_p view, 
                                                            OBIDMS_column_p column_p, 
                                                            index_t line_nb, 
                                                            const char* element_name,
                                                            int offset)

    const uint8_t* obi_get_qual_int_with_elt_name_and_col_p_in_view(Obiview_p view, 
                                                                    OBIDMS_column_p column_p, 
                                                                    index_t line_nb, 
                                                                    const char* element_name, 
                                                                    int* value_length)


    # OBI_STR
    int obi_set_str_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                    OBIDMS_column_p column_p, 
                                                    index_t line_nb, 
                                                    const_char_p element_name, 
                                                    const_char_p value)

    int obi_set_str_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                   OBIDMS_column_p column_p, 
                                                   index_t line_nb, 
                                                   index_t element_idx, 
                                                   const_char_p value)
    
    const_char_p obi_get_str_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                           OBIDMS_column_p column_p, 
                                                           index_t line_nb, 
                                                           const_char_p element_name)
    
    const_char_p obi_get_str_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                            OBIDMS_column_p column_p, 
                                                            index_t line_nb, 
                                                            index_t element_idx)


    # OBI_SEQ
    int obi_set_seq_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                    OBIDMS_column_p column_p, 
                                                    index_t line_nb, 
                                                    const_char_p element_name, 
                                                    const_char_p value)

    int obi_set_seq_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                   OBIDMS_column_p column_p, 
                                                   index_t line_nb, 
                                                   index_t element_idx, 
                                                   const_char_p value)
    
    char* obi_get_seq_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                      OBIDMS_column_p column_p, 
                                                      index_t line_nb, 
                                                      const_char_p element_name)
    
    char* obi_get_seq_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                     OBIDMS_column_p column_p, 
                                                     index_t line_nb, 
                                                     index_t element_idx)

    
    # OBI_IDX
    int obi_set_index_with_elt_name_and_col_p_in_view(Obiview_p view, 
                                                     OBIDMS_column_p column_p, 
                                                     index_t line_nb, 
                                                     const_char_p element_name, 
                                                     index_t value)

    int obi_set_index_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                     OBIDMS_column_p column_p, 
                                                     index_t line_nb, 
                                                     index_t element_idx, 
                                                     index_t value)
    
    index_t obi_get_index_with_elt_name_and_col_p_in_view(Obiview_p view,
                                                          OBIDMS_column_p column_p, 
                                                          index_t line_nb, 
                                                          const_char_p element_name)

    index_t obi_get_index_with_elt_idx_and_col_p_in_view(Obiview_p view,
                                                         OBIDMS_column_p column_p, 
                                                         index_t line_nb, 
                                                         index_t element_idx)

    # ARRAY
    int obi_set_array_with_col_p_in_view(Obiview_p view, 
                                         OBIDMS_column_p column, 
                                         index_t line_nb, 
                                         const void* value, 
                                         uint8_t elt_size, 
                                         int32_t value_length)
    
    const void* obi_get_array_with_col_p_in_view(Obiview_p view, 
                                                 OBIDMS_column_p column, 
                                                 index_t line_nb, 
                                                 int32_t* value_length_p)

    int obi_set_array_with_col_name_in_view(Obiview_p view, 
                                            const char* column_name, 
                                            index_t line_nb, 
                                            const void* value, 
                                            uint8_t elt_size, 
                                            int32_t value_length)

    const void* obi_get_array_with_col_name_in_view(Obiview_p view, 
                                                    const char* column_name, 
                                                    index_t line_nb, 
                                                    int32_t* value_length_p)




#cython: language_level=3

'''
Created on 25 mars 2016

@author: coissac
'''

from obitools3.utils cimport __etag__
import re


__ret__  = re.compile(b'''(([^ ]+)=('[^']*'|"[^"]*"|[^;]+); *)+?''')


cpdef tuple parseHeader(bytes header, bytes nastring=b"NA"):
    cdef list  m 
    cdef dict  tags 
    cdef bytes definition
    cdef bytes ident
    cdef bytes second
    
    m=header[1:-1].split(maxsplit=1)
   
    ident=m[0]
    if len(ident)>1 and ident[-2:-1] == b';':
        ident = ident[:-1]
    
    if len(m)==1:
        tags={}
        definition=b'' 
    else:
        second=m[1]
        m = __ret__.findall(second)
        
        if m:
            tags = dict([(a[1],__etag__(a[2], nastring=nastring)) for a in m])
            definition = second.split(m[-1][0],1)[1].strip()
        else:
            tags = {}
            definition = second.strip() 
        
    return ident,tags,definition
    




#cython: language_level=3

from cpython cimport array

from .solexapairend import iterOnAligment
from .shifted_ali cimport Ali_shifted

from obitools3.dms.capi.obiview cimport Obiview_p, QUALITY_COLUMN, NUC_SEQUENCE_COLUMN, \
                                        REVERSE_SEQUENCE_COLUMN, REVERSE_QUALITY_COLUMN, \
                                        obi_set_qual_int_with_elt_idx_and_col_p_in_view, \
                                        obi_set_str_with_elt_idx_and_col_p_in_view
                                        
from obitools3.dms.capi.obidmscolumn cimport OBIDMS_column_p

from obitools3.dms.view.view cimport View
from obitools3.dms.column.column cimport Column

from math import log10


cdef class IterOnConsensus:

    cdef object _ali
    cdef int __seqASingle
    cdef int __seqBSingle
    cdef int __seqABMatch
    cdef int __seqAMismatch
    cdef int __seqBMismatch
    cdef int __seqAInsertion
    cdef int __seqBInsertion
    cdef int __seqADeletion
    cdef int __seqBDeletion
    cdef object __ioa
    cdef bint __firstSeqB

    def __cinit__(self,ali):
        self._ali=ali
        self.__seqASingle=0
        self.__seqBSingle=0
        self.__seqABMatch=0
        self.__seqAMismatch=0
        self.__seqBMismatch=0
        self.__seqAInsertion=0
        self.__seqBInsertion=0
        self.__seqADeletion=0
        self.__seqBDeletion=0

        self.__ioa = iterOnAligment(self._ali)
        self.__firstSeqB=False

    def get_seqASingle(self):
        return self.__seqASingle


    def get_seqBSingle(self):
        return self.__seqBSingle


    def get_seqABMatch(self):
        return self.__seqABMatch


    def get_seqAMismatch(self):
        return self.__seqAMismatch


    def get_seqBMismatch(self):
        return self.__seqBMismatch


    def get_seqAInsertion(self):
        return self.__seqAInsertion


    def get_seqBInsertion(self):
        return self.__seqBInsertion


    def get_seqADeletion(self):
        return self.__seqADeletion


    def get_seqBDeletion(self):
        return self.__seqBDeletion

    def __next__(self):
        cdef bytes snuc0
        cdef bytes snuc1
        cdef char* nuc0
        cdef char* nuc1
        cdef char* dash=b"-"
        cdef double score0
        cdef double score1
        cdef double h0
        cdef double h1

        while(1):
            
            snuc0,score0,snuc1,score1 = self.__ioa.__next__()
            nuc0=snuc0
            nuc1=snuc1
            if nuc0[0]==nuc1[0]:
                if nuc1[0]!=dash[0]:
                    self.__firstSeqB=True
                    self.__seqABMatch+=1
                    self.__seqBSingle=0
                    return (nuc0,score0*score1)
            else:
                h0 = score0 * (1-score1/3)
                h1 = score1 * (1-score0/3)
                if h0 < h1:

                    if nuc0[0]!=dash[0]:
                        self.__seqBSingle=0
                        if nuc1[0]==dash[0]:
                            if self.__firstSeqB:
                                self.__seqAInsertion+=1
                            else:
                                self.__seqASingle+=1
                        else:
                            self.__firstSeqB=True
                            self.__seqAMismatch+=1
                        if score1==1.0:    # Free end gap
                            return (nuc0,score0)
                        else:
                            return (nuc0,h0)
                    else:
                        self.__seqADeletion+=1
                else:
                    if nuc1[0]!=dash[0]:
                        self.__firstSeqB=True
                        if nuc0[0]==dash[0]:
                            self.__seqBInsertion+=1
                            self.__seqBSingle+=1
                        else:
                            self.__seqBMismatch+=1
                            self.__seqBSingle=0
                        if score0==1.0:    # Free end gap
                            return (nuc1,score1)
                        else:
                            return (nuc1,h1)
                    else:
                        self.__seqBSingle=0
                        self.__seqBDeletion+=1


    def __iter__(self):
        return self

    seqASingle = property(get_seqASingle, None, None, "direct's docstring")
    seqBSingle = property(get_seqBSingle, None, None, "reverse's docstring")
    seqABMatch = property(get_seqABMatch, None, None, "idem's docstring")
    seqAMismatch = property(get_seqAMismatch, None, None, "mismatchdirect's docstring")
    seqBMismatch = property(get_seqBMismatch, None, None, "mismatchreverse's docstring")
    seqAInsertion = property(get_seqAInsertion, None, None, "insertdirect's docstring")
    seqBInsertion = property(get_seqBInsertion, None, None, "insertreverse's docstring")
    seqADeletion = property(get_seqADeletion, None, None, "deletedirect's docstring")
    seqBDeletion = property(get_seqBDeletion, None, None, "deletereverse's docstring")


def buildConsensus(ali, seq, ref_tags=None):
    cdef list quality
    cdef char   aseq[1000]
    cdef int i=0
    cdef int j=0
    cdef char* cnuc
    cdef bytes nuc
    cdef double score
    cdef bytes sseq
    cdef double p
    cdef OBIDMS_column_p col_p
    cdef Obiview_p view_p
    cdef View view
    cdef Column column
    
    quality = []
        
    if type(ali) == Ali_shifted:
        view = seq.view
        view_p = view.pointer()
        column = view[QUALITY_COLUMN]
        col_p = column.pointer()
        # doesn't work because uint8_t* are forced into bytes by cython (nothing read/kept beyond 0 values)
        #obi_set_qual_int_with_elt_idx_and_col_p_in_view(view_p, col_p, seq.index, 0, ali.consensus_qual, ali.consensus_len)
        seq.set(ref_tags.id+b"_CONS", ali.consensus_seq, quality=ali.consensus_qual)
        seq[b"seq_length"] = ali.consensus_len
        seq[b"overlap_length"] = ali.overlap_len
        seq[b'score_norm']=round(float(ali.score)/ali.overlap_len, 3)
        seq[b'shift']=ali.shift
    else:
        if len(ali[0])>999:   # TODO why?
            raise AssertionError,"Too long alignment"
    
        ic=IterOnConsensus(ali)
    
        for nuc,score in ic:
            cnuc=nuc
            quality.append(score)
            aseq[i]=cnuc[0]
            i+=1
    
        aseq[i]=0
    
        sseq=aseq
    
        # Reconvert quality from array of probabilities to array of raw quality values
        i=0
        for p in quality:
            quality[i] = min(42, round(-10*log10(p)))
            i+=1
        
        seq.set(ali[0].wrapped.id+b"_CONS", sseq, quality=quality)
        
        if hasattr(ali, "counter"):
            seq[b'alignement_id']=ali.counter
        seq[b'seq_a_single']=ic.seqASingle
        seq[b'seq_b_single']=ic.seqBSingle
        seq[b'seq_ab_match']=ic.seqABMatch
        seq[b'seq_a_mismatch']=ic.seqAMismatch
        seq[b'seq_b_mismatch']=ic.seqBMismatch
        seq[b'seq_a_insertion']=ic.seqAInsertion
        seq[b'seq_b_insertion']=ic.seqBInsertion-ic.seqBSingle
        seq[b'seq_a_deletion']=ic.seqADeletion
        seq[b'seq_b_deletion']=ic.seqBDeletion
        seq[b'ali_length']=len(seq)-ic.seqASingle-ic.seqBSingle
        if seq[b'ali_length']>0:
            seq[b'score_norm']=float(ali.score)/seq[b'ali_length']
        
        ref_tags = ali[0].wrapped

    if hasattr(ali, "direction"):
        seq[b'ali_direction']=ali.direction
    seq[b'score']=ali.score
    seq[b'mode']=b'alignment'

    for tag in ref_tags:
        if tag != REVERSE_SEQUENCE_COLUMN and tag != REVERSE_QUALITY_COLUMN and \
            tag != NUC_SEQUENCE_COLUMN and tag != QUALITY_COLUMN:
            seq[tag] = ref_tags[tag]

    return seq


def buildJoinedSequence(ali, reverse, seq, forward=None):
    if forward is not None:
        forward = forward
    else:
        forward = ali[0].wrapped
    s = forward.seq + reverse.seq     
    quality = forward.quality
    quality.extend(reverse.quality)
    seq.set(forward.id +b"_PairedEnd", s, definition=forward.definition, quality=quality)
    seq[b"score"]=ali.score
    if len(ali.direction) > 0:
        seq[b"ali_direction"]=ali.direction
    else:
        seq[b"ali_direction"]=None
    seq[b"mode"]=b"joined"
    seq[b"pairedend_limit"]=len(forward)
    seq[b"seq_length"] = ali.consensus_len
    seq[b"overlap_length"] = ali.overlap_len
    if ali.overlap_len > 0:
        seq[b'score_norm']=round(float(ali.score)/ali.overlap_len, 3)
    else:
        seq[b"score_norm"]=0.0
  
    for tag in forward:
        if tag != REVERSE_SEQUENCE_COLUMN and tag != REVERSE_QUALITY_COLUMN and \
            tag != NUC_SEQUENCE_COLUMN and tag != QUALITY_COLUMN:
            seq[tag] = forward[tag]
    return seq




#cython: language_level=3

from obitools3.dms.capi.obitypes cimport is_a_DNA_seq, \
                                         OBI_VOID, \
                                         OBI_BOOL, \
                                         OBI_CHAR, \
                                         OBI_FLOAT, \
                                         OBI_INT, \
                                         OBI_QUAL, \
                                         OBI_SEQ, \
                                         OBI_STR, \
                                         index_t, \
                                         OBI_INT_MAX

from obitools3.dms.capi.obierrno cimport OBI_LINE_IDX_ERROR, \
                                         OBI_ELT_IDX_ERROR, \
                                         obi_errno

from obitools3.files.uncompress cimport CompressedFile

import re
import mmap
import os
import glob
import gzip


cpdef bytes format_uniq_pattern(bytes format):
    if format == b"fasta":
        return b"\n>"
    elif format == b"fastq":
        return b"\n\+\n"
    elif format == b"ngsfilter" or format == b"tabular":
        return b"\n"
    elif format == b"genbank" or format == b"embl":
        return b"\n//"
    elif format == b"ecopcr":
        return b"\n[^#]"
    else:
        return None


cpdef int count_entries(file, bytes format, bint header):
    
    try:
        sep = format_uniq_pattern(format)
        if sep is None:
            return -1
        sep = re.compile(sep)

        if type(file) == bytes and (format == b'genbank' or format == b'embl'): # file is actually a directory with multiple files
            files = []
            if format == b'embl':
                extensions = [b"*.dat"]
            elif format == b"genbank":
                extensions = [b"*.gbff"]
            
            for ext in extensions:
                for filename in glob.glob(os.path.join(file, ext)):
                    #if filename[:-3] == ".gz":
                    #    files.append(gzip.open(filename, "rb"))
                    #else:
                    files.append(open(filename, "rb"))
        else:
            files = [file]
        
        if len(files)==0:
            return -1
        
        total_count = 0
        for f in files:
            if type(f) == CompressedFile and f.compressed:
                return -1
            mmapped_file = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
            total_count += len(re.findall(sep, mmapped_file))
            if format != b"ngsfilter" and format != b"tabular" and format != b"embl" and format != b"genbank" and format != b"fastq":
                total_count += 1 # adding +1 for 1st entry because separators include \n (ngsfilter and tabular already count one more because of last \n)
            if format == b"tabular" and header: # not counting header as an entry
                total_count -= 1
            
    except:
        if len(files) > 1:
            for file in files:
                file.close()
        return -1
    
    if len(files) > 1:
        for f in files:
            f.close()
    
    return total_count


# TODO RollbackException?
cdef obi_errno_to_exception(index_t line_nb=-1, object elt_id=None, str error_message=None) :
    global obi_errno
    if obi_errno > 0 :
        if obi_errno == OBI_LINE_IDX_ERROR :
            raise IndexError(line_nb, None or error_message)
        elif obi_errno == OBI_ELT_IDX_ERROR :
            raise IndexError(elt_id, None or error_message)
        else :
            raise Exception(None or error_message)


cdef bytes str2bytes(str string):
    """
    Short cut to convert ascii encoded python string (str) to bytes 
    which can be easily converted to C-strings. 
    
        @param string: the python string to be converted.
        @type string: str
        @return a transcoded string
        @rtype: bytes 
    """
    return string.encode('ascii')

cdef str bytes2str(bytes string):
    """
    Short cut to convert bytes (C-strings) to ascii encoded python string (str).
    
        @param string: the binary (C-string) string to be converted.
        @type string: bytes
        @return an ascii transcoded string
        @rtype: str 
    """
    return string.decode('ascii')

cdef bytes tobytes(object string):
    """
    Short cut to convert ascii encoded string (str or bytes) to bytes 
    which can be easily converted to C-strings. 
    
        @param string: the python string to be converted.
        @type string: bytes or str
        @return a transcoded string
        @rtype: bytes 
    """
    if string is None:
        return None
    if isinstance(string, bytes):
        return string
    return str2bytes(string)


cdef str tostr(object string):
    """
    Short cut to convert ascii encoded string (str or bytes) to bytes 
    which can be easily converted to C-strings. 
    
        @param string: the python string to be converted.
        @type string: bytes or str
        @return a transcoded string
        @rtype: bytes 
    """
    if isinstance(string, str):
        return string
    return bytes2str(string)


cdef object bytes2str_object(object value):  # Only works if complex types are dict or list
    if isinstance(value, dict):
        items = [(k,v) for k,v in value.items()]
        for k, v in items:
            if isinstance(v, list) or isinstance(v, dict):
                value[k] = bytes2str_object(v)
            else:
                if type(v) == bytes:
                    value[k] = bytes2str(v)
            if type(k) == bytes:
                value[bytes2str(k)] = value.pop(k)
    elif isinstance(value, list) or isinstance(value, tuple):
        if isinstance(value, tuple):
            value = list(value)
        for i in range(len(value)):
            if isinstance(value[i], list) or isinstance(value[i], dict):
                value[i] = bytes2str_object(value[i])
            if type(value[i]) == bytes:
                value[i] = bytes2str(value[i])
    elif type(value) == bytes:
        value = bytes2str(value)
    return value


cdef object str2bytes_object(object value):  # Only works if complex types are dict or list
    if isinstance(value, dict):
        items = [(k,v) for k,v in value.items()]
        for k, v in items:
            if isinstance(v, list) or isinstance(v, dict):
                value[k] = str2bytes_object(v)
            else:
                if type(v) == str:
                    value[k] = str2bytes(v)
            if type(k) == str:
                value[str2bytes(k)] = value.pop(k)
    elif isinstance(value, list):
        for i in range(len(value)):
            if isinstance(value[i], list) or isinstance(value[i], dict):
                value[i] = str2bytes_object(value[i])
            if type(value[i]) == str:
                value[i] = str2bytes(value[i])
    elif type(value) == str:
        value = str2bytes(value)
    return value


cdef object clean_empty_values_from_object(object value, exclude=[]):    # Only works if complex types are dict or list
    if isinstance(value, dict):
        items = [(k,v) for k,v in value.items()]
        for k, v in items:
            if isinstance(v, list) or isinstance(v, dict):
                value[k] = clean_empty_values_from_object(v)
            if (k not in exclude) and (v is None or (hasattr(v, '__len__') and len(v) == 0)):
                value.pop(k)
    elif isinstance(value, list):
        to_remove=[]
        for i in range(len(value)):
            if isinstance(value[i], list) or isinstance(value[i], dict):
                value[i] = clean_empty_values_from_object(value[i])
            if value[i] is None or (hasattr(value[i], '__len__') and len(value[i]) == 0):
                to_remove.append(value[i])
        for v in to_remove:
            value.remove(v)
    return value


cdef obitype_t get_obitype_single_value(object value) :

    cdef type       value_type
    cdef obitype_t  value_obitype
    
    if value is None :
        return OBI_VOID
    
    value_type = type(value)
    value_obitype = OBI_VOID
                
    if value_type == int :
        value_obitype = OBI_INT
    elif value_type == float :
        value_obitype = OBI_FLOAT
    elif value_type == bool :
        value_obitype = OBI_BOOL        
    elif value_type == str or value_type == bytes :
        if is_a_DNA_seq(tobytes(value)): #or value_type == Nuc_Seq or value_type == Nuc_Seq_Stored:  # TODO discuss
            value_obitype = OBI_SEQ
        elif len(value) == 1 :
            value_obitype = OBI_CHAR
        elif (len(value) > 1) :
            value_obitype = OBI_STR
    else :
        value_obitype = OBI_VOID
    
    return value_obitype


cdef obitype_t update_obitype(obitype_t obitype, object new_value) :
    
    cdef type new_type
    
    new_type = type(new_value)
    
    #if new_type == NoneType:  # doesn't work because Cython sucks
    if new_value == None or new_type==list or new_type==dict or new_type==tuple:
        return obitype
    
    # TODO BOOL to INT/FLOAT
    if new_type == str or new_type == bytes :
        if obitype == OBI_SEQ and is_a_DNA_seq(tobytes(new_value)) :
            pass
        else :
            return OBI_STR
    elif obitype == OBI_INT :
        if new_type == float or new_value > OBI_INT_MAX :
            return OBI_FLOAT
        
    return obitype


cdef obitype_t get_obitype_iterable_value(object value, type t) :
    
    cdef obitype_t value_obitype
    
    value_obitype = OBI_VOID
    
    if t == dict:
        for k in value :
            if value_obitype == OBI_VOID :
                value_obitype = get_obitype_single_value(value[k])
            else :
                value_obitype = update_obitype(value_obitype, value[k])
    
    elif t == list or t == tuple:
        for v in value :
            if value_obitype == OBI_VOID :
                value_obitype = get_obitype_single_value(v)
            else :
                value_obitype = update_obitype(value_obitype, v)
    
    return value_obitype


cdef obitype_t get_obitype(object value) :
    
    t = type(value)
    if t == dict or t == list or t == tuple :
        return get_obitype_iterable_value(value, t)
    
    else :
        return get_obitype_single_value(value)


__re_int__      = re.compile(b"^[+-]?[0-9]+$")
__re_float__    = re.compile(b"^[+-]?[0-9]+(\.[0-9]*)?([eE][+-]?[0-9]+)?$")
__re_str__      = re.compile(b"""^"[^"]*"|'[^']*'$""")
__re_dict__     = re.compile(b"""^\{\ *
                                   (
                                       ("[^"]*"|'[^']*')
                                        \ *:\ *
                                       ([^,}]+|
                                        "[^"]*"|
                                        '[^']*'
                                       )
                                   )?
                                   (\ *,\ *
                                       ("[^"]*"|'[^']*')
                                        \ *:\ *
                                       ([^,}]+|
                                        "[^"]*"|
                                        '[^']*'
                                       )
                                    )*\ *\}$""", re.VERBOSE)

__re_val__ = re.compile(b"""(("[^"]*"|'[^']*') *: *([^,}]+|"[^"]*"|'[^']*') *[,}] *)""")

cdef object __etag__(bytes x, bytes nastring=b"NA"):
    cdef list elements
    cdef tuple i
    
    if x == nastring:
        v = None
    elif __re_int__.match(x):
        v=int(x)
    elif __re_float__.match(x):
        v=float(x)
    elif __re_str__.match(x):
        v=x[1:-1]
    elif x==b'None':
        v=None
    elif x==b'False':
        v=False
    elif x==b'True':
        v=True
    elif __re_dict__.match(x):
        elements=__re_val__.findall(x)
        v=dict([(i[1][1:-1],__etag__(i[2])) for i in elements])
    else:
        v=x
    return v



#cython: language_level=3

from .obitypes cimport const_char_p, \
                       obiversion_t

cdef extern from "obidms.h" nogil:

    struct OBIDMS_infos_t :
        bint                little_endian
        size_t              file_size
        size_t              used_size
        bint                working
        const_char_p        comments

    ctypedef OBIDMS_infos_t* OBIDMS_infos_p

    
    struct OBIDMS_t:
        const_char_p        dms_name
        const_char_p        directory_path
        OBIDMS_infos_p      infos

    ctypedef OBIDMS_t* OBIDMS_p 

    int obi_dms_is_clean(OBIDMS_p dms)
    int obi_clean_dms(const_char_p dms_path)
    OBIDMS_p obi_dms(const_char_p dms_name)
    OBIDMS_p obi_open_dms(const_char_p dms_path, bint cleaning)
    OBIDMS_p obi_test_open_dms(const_char_p dms_path)
    OBIDMS_p obi_create_dms(const_char_p dms_path)
    int obi_dms_exists(const char* dms_path)
    int obi_dms_write_comments(OBIDMS_p dms, const char* comments)
    int obi_dms_add_comment(OBIDMS_p dms, const char* key, const char* value)
    int obi_close_dms(OBIDMS_p dms, bint force)
    char* obi_dms_get_dms_path(OBIDMS_p dms)
    char* obi_dms_get_full_path(OBIDMS_p dms, const_char_p path_name)
    char* obi_dms_formatted_infos(OBIDMS_p dms, bint detailed)
    void obi_close_atexit()
    
    obiversion_t obi_import_column(const char* dms_path_1, const char* dms_path_2, const char* column_name, obiversion_t version_number)
    int obi_import_view(const char* dms_path_1, const char* dms_path_2, const char* view_name_1, const char* view_name_2)



#cython: language_level=3

from obitools3.apps.optiongroups import addMinimalInputOption
from obitools3.uri.decode import open_uri
from obitools3.dms import DMS
from obitools3.dms.view import View
from obitools3.utils cimport bytes2str

 
__title__="Command line histories and view history graphs"
 

def addOptions(parser):

    addMinimalInputOption(parser)
 
    group=parser.add_argument_group('obi history specific options')
 
    group.add_argument('--bash', '-b',
                     action="store_const", dest="history:format",
                     default="bash",
                     const="bash",
                     help="Print history in bash format")

    group.add_argument('--dot', '-d',
                     action="store_const", dest="history:format",
                     default="bash",
                     const="dot",
                     help="Print history in DOT format (default: bash format)")

    group.add_argument('--ascii', '-a',
                     action="store_const", dest="history:format",
                     default="bash",
                     const="ascii",
                     help="Print history in ASCII format (only for views; default: bash format)")
 

def run(config):
    
    cdef object entries
    
    DMS.obi_atexit()
    
    input = open_uri(config['obi']['inputURI'])
       
    entries = input[1]
    
    if config['history']['format'] == "bash" :
        print(bytes2str(entries.bash_history))
    elif config['history']['format'] == "dot" :
        print(bytes2str(entries.dot_history_graph))
    elif config['history']['format'] == "ascii" :
        if isinstance(entries, View):
            print(bytes2str(entries.ascii_history))
        else:
            raise Exception("ASCII history only available for views")
    
    input[0].close(force=True)



#cython: language_level=3

'''
Created on feb 20th 2018

@author: cmercier
'''

import types
from obitools3.utils cimport __etag__
from obitools3.utils cimport str2bytes

def tabIterator(lineiterator, 
                bint header = False,
                bytes sep = None,
                bytes dec = b".",          # TODO don't know how to use this to parse
                bint stripwhite=True,
                bint blanklineskip=True,
                bytes commentchar=b"#",
                bytes nastring=b"NA",
                int skip=0,
                only=None,
                firstline=None,
                int buffersize=100000000
                ):
    
    cdef LineBuffer lb
    cdef int        lines_to_skip, ionly, read
    cdef list       data
    cdef dict       view_line
    cdef list       keys
    cdef list       key_types
        
    keys = []
    key_types = []
    skipped = 0
    read = 0
    
    if only is None:
        ionly = -1
    else:
        ionly = int(only)
        
    if isinstance(lineiterator, (str, bytes)):
        lineiterator=uopen(lineiterator)        
    if isinstance(lineiterator, LineBuffer):
        iterator = iter(lineiterator)
    else:
        if hasattr(lineiterator, "readlines"):
            iterator = iter(LineBuffer(lineiterator, buffersize))
        elif hasattr(lineiterator, '__next__'):
            iterator = lineiterator
        else:
            raise Exception("Invalid line iterator")
    
    if firstline is None:
        line = next(iterator)
    else:
        line = firstline       
    
    while True:
        
        if (not line.strip() and blanklineskip) or line[:1] == commentchar:
            line = next(iterator)
        
        if ionly >= 0 and read >= ionly:
            break

        if not keys:
            if header:
                # TODO read types eventually
                keys = line.split(sep)
                keys = [x.strip() for x in keys]
                line = next(iterator)
                continue
            else:
                # TODO ??? default column names? like R?
                keys = [str2bytes(str(i)) for i in range(len(line.split(sep)))]
                
        while skipped < skip :
            line = next(iterator)
            skipped += 1

        view_line = {}
        
        # Parse
        data = line.split(sep)

        if stripwhite or key_types:
            data = [x.strip() for x in data]
        
        for i in range(len(data)):
            if key_types:  # TODO handle None when key types are actually read
                view_line[keys[i]] = key_types[i](data[i])
            else:
                view_line[keys[i]] = __etag__(data[i], nastring=nastring)
        
        yield view_line
        
        read+=1
        
        try:
            line = next(iterator)
        except StopIteration:
            return 
    
        
        


#cython: language_level=3

from obitools3.uri.decode import open_uri
from obitools3.apps.config import logger
from obitools3.dms import DMS
from obitools3.dms.taxo.taxo cimport Taxonomy
from obitools3.apps.optiongroups import addMinimalInputOption
from obitools3.utils cimport tostr, bytes2str_object

 
__title__="Print a preview of a DMS, view, column...."


def addOptions(parser):    
    addMinimalInputOption(parser)
    group = parser.add_argument_group('obi ls specific options')

    group.add_argument('-l',
                     action="store_true", dest="ls:longformat",
                     default=False,
                     help="Detailed list in long format with all metadata.")


def run(config):

    DMS.obi_atexit()
    
    logger("info", "obi ls")

    # Open the input
    input = open_uri(config['obi']['inputURI'])
    if input is None:
        raise Exception("Could not read input")
    
    # Print representation
    if config['ls']['longformat']:
        print(input[1].repr_longformat())
    else:
        print(repr(input[1]))
    
    input[0].close(force=True)



#cython: language_level=3

'''
Created on march 8th 2018

@author: cmercier
'''

from .tab import tabIterator
import types


def ngsfilterIterator(lineiterator, 
                      bytes sep = None,
                      bytes dec = b".",
                      bint stripwhite=True,
                      bint blanklineskip=True,
                      bytes commentchar=b"#",
                      bytes nastring=b"NA",
                      int skip=0,
                      only=None,
                      firstline=None,
                      int buffersize=100000000
                      ):
    
    cdef list  all_lines
    cdef bytes header 
    cdef bytes out_sep
        
    out_sep = b"\t"

    if isinstance(lineiterator, (str, bytes)):
        lineiterator=uopen(lineiterator)        
    if isinstance(lineiterator, LineBuffer):
        iterator = iter(lineiterator)
    else:
        if hasattr(lineiterator, "readlines"):
            iterator = iter(LineBuffer(lineiterator, buffersize))
        elif hasattr(lineiterator, '__next__'):
            iterator = lineiterator
        else:
            raise Exception("Invalid line iterator")
    
    all_lines = [line for line in iterator]
    new_lines = []
        
    if firstline is not None:
        all_lines.insert(0, firstline)
        
    # Insert header for column names
    column_names = [b"experiment", b"sample", b"forward_tag", b"reverse_tag", b"forward_primer", b"reverse_primer"] #,b"additional_info"]
    header = out_sep.join(column_names)
    
    new_lines.append(header)
        
    for line in all_lines:
        split_line = line.split(maxsplit=5)
        if split_line:
            tags = split_line.pop(2)
            tags = tags.split(b":")
            for t_idx in range(len(tags)):
                if tags[t_idx]==b"-" or tags[t_idx]==b"None" or tags[t_idx]==b"":
                    tags[t_idx] = nastring
            if len(tags) == 1:          # Forward and reverse tags are the same
                tags.append(tags[0])
            split_line.insert(2, tags[0])
            split_line.insert(3, tags[1])
            new_lines.append(out_sep.join(split_line[0:6]))
        
    return tabIterator(iter(new_lines),
                       header = True,
                       sep = out_sep,
                       dec = dec,
                       stripwhite = stripwhite,
                       blanklineskip = blanklineskip,
                       commentchar = commentchar,
                       nastring = nastring,
                       skip = skip,
                       only = only,
                       firstline = None)




from ..utils cimport bytes2str
from .header cimport HeaderFormat
from cython.view cimport array as cvarray

cdef class FastqFormat:

    cdef HeaderFormat headerFormatter
    
    cdef size_t  sequenceBufferLength
    cdef char*   sequenceBuffer



#cython: language_level=3

from obitools3.apps.progress cimport ProgressBar  # TODO I absolutely don't understand why it doesn't work without that line
from obitools3.dms.view import View, Line_selection
from obitools3.dms.view.typed_view.view_NUC_SEQS import View_NUC_SEQS
from obitools3.dms import DMS
from obitools3.dms.column import Column
from obitools3.dms.taxo import Taxonomy
from obitools3.utils cimport str2bytes
from obitools3.dms.capi.obitypes cimport OBI_INT, \
                                         OBI_FLOAT, \
                                         OBI_BOOL, \
                                         OBI_CHAR, \
                                         OBI_STR, \
                                         OBI_SEQ
                                         
from obitools3.dms.capi.obiview cimport NUC_SEQUENCE_COLUMN, \
                                        ID_COLUMN, \
                                        DEFINITION_COLUMN, \
                                        QUALITY_COLUMN, \
                                        COUNT_COLUMN

import shutil
import string
import random
import sys
from cpython.exc cimport PyErr_CheckSignals


VIEW_TYPES = [b"", b"NUC_SEQS_VIEW"]
COL_TYPES = [OBI_INT, OBI_FLOAT, OBI_BOOL, OBI_CHAR, OBI_STR, OBI_SEQ]
SPECIAL_COLUMNS = [NUC_SEQUENCE_COLUMN, ID_COLUMN, DEFINITION_COLUMN, QUALITY_COLUMN]

#TAXDUMP = "" TODO path=?
TAXTEST = b"taxtest"
 
NAME_MAX_LEN = 200
COL_COMMENTS_MAX_LEN = 2048
MAX_INT = 2147483647    # used to generate random float values
 
 
__title__="Test if the obitools are working properly"


default_config = {
                 }


def test_taxo(config, infos):
    tax1 = Taxonomy.open_taxdump(infos['dms'], config['obi']['taxo'])
    tax1.write(TAXTEST)
    tax2 = Taxonomy.open(infos['dms'], TAXTEST)
    assert len(tax1) == len(tax2), "Length of written taxonomy != length of read taxdump : "+str(len(tax2))+" != "+str(len(tax1))
    
    i = 0
    for x in range(config['test']['nbtests']):
        idx = random.randint(0, len(tax1)-1)
        t1 = tax1.get_taxon_by_idx(idx)
        taxid1 = t1.taxid
        t2 = tax2.get_taxon_by_idx(idx)
        taxid2 = t2.taxid
        assert t1 == t2, "Taxon gotten from written taxonomy with index != taxon read from taxdump : "+str(t2)+" != "+str(t1)
        t1 = tax1[taxid1]
        t2 = tax2[taxid2]
        assert t1 == t2, "Taxon gotten from written taxonomy with taxid != taxon read from taxdump : "+str(t2)+" != "+str(t1)
        i+=1
        if (i%(config['test']['nbtests']/10)) == 0 :
            print("Testing taxonomy functions......"+str(i*100/config['test']['nbtests'])+"%")

    tax1.close()
    tax2.close()


def random_length(max_len):
    return random.randint(1, max_len)


def random_bool(config):
    return random.choice([True, False])


def random_bool_tuples(config):
    l=[]
    for i in range(random.randint(1, config['test']['tuplemaxlen'])) :
        l.append(random.choice([None, random_bool(config)]))
    return tuple(l)


def random_char(config):
    return str2bytes(random.choice(string.ascii_lowercase))


def random_char_tuples(config):
    l=[]
    for i in range(random.randint(1, config['test']['tuplemaxlen'])) :
        l.append(random.choice([None, random_char(config)]))
    return tuple(l)


def random_float(config):
    return random.randint(0, MAX_INT) + random.random()


def random_float_tuples(config):
    l=[]
    for i in range(random.randint(1, config['test']['tuplemaxlen'])) :
        l.append(random.choice([None, random_float(config)]))
    return tuple(l)


def random_int(config):
    return random.randint(0, config['test']['maxlinenb'])


def random_int_tuples(config):
    l=[]
    for i in range(random.randint(1, config['test']['tuplemaxlen'])) :
        l.append(random.choice([None, random_int(config)]))
    return tuple(l)


def random_seq(config):
    return str2bytes(''.join(random.choice(['a','t','g','c']) for i in range(random_length(config['test']['seqmaxlen']))))


def random_seq_tuples(config):
    l=[]
    for i in range(random.randint(1, config['test']['tuplemaxlen'])) :
        l.append(random.choice([None, random_seq(config)]))
    return tuple(l)


def random_bytes(config):
    return random_bytes_with_max_len(config['test']['strmaxlen'])


def random_bytes_tuples(config):
    l=[]
    for i in range(random.randint(1, config['test']['tuplemaxlen'])) :
        l.append(random.choice([None, random_bytes(config)]))
    return tuple(l)


def random_str_with_max_len(max_len):
    return ''.join(random.choice(string.ascii_lowercase) for i in range(random_length(max_len)))


def random_bytes_with_max_len(max_len):
    return str2bytes(random_str_with_max_len(max_len))


RANDOM_FUNCTIONS = [random_bool, random_char, random_bytes, random_float, random_int]
def random_comments(config):
    comments = {}
    for i in range(random_length(1000)):
        to_add = {random_bytes(config): random.choice(RANDOM_FUNCTIONS)(config)}
        if len(str(comments)) + len(str(to_add)) >= COL_COMMENTS_MAX_LEN:
            return comments
        else:
            comments.update(to_add)
    return comments
    
 
def random_column(infos):
    return random.choice(sorted(list(infos['view'].keys())))
 
 
def random_unique_name(infos):
    name = b""
    while name == b"" or name in infos['unique_names'] :
        name = random_bytes_with_max_len(NAME_MAX_LEN)
    infos['unique_names'].append(name)
    return name
 
 
def random_unique_element_name(config, infos):
    name = b""
    while name == b"" or name in infos['unique_names'] :
        name = random_bytes_with_max_len(config['test']['elt_name_max_len'])
    infos['unique_names'].append(name)
    return name
 
 
def print_test(config, sentence):
    if config['test']['verbose'] :
        print(sentence)
 
 
def test_set_and_get(config, infos):
    print_test(config, ">>> Set and get test")
    col_name = random_column(infos)
    col = infos['view'][col_name]
    element_names = col.elements_names
    data_type = col.data_type
    if data_type == b"OBI_QUAL" :
        print_test(config, "-")
        return
    idx = random_int(config)
    value = random.choice([None, infos['random_generator'][(data_type, col.tuples)](config)])
    if col.nb_elements_per_line > 1 :
        elt = random.choice(element_names)
        col[idx][elt] = value
        assert col[idx][elt] == value, "Column: "+repr(col)+"\nSet value != gotten value "+str(value)+" != "+str(col[idx][elt])
    elif col.tuples:
        col[idx] = value
        if value is None:
            totest = None
        else:
            totest = []
            for e in value:
                if e is not None and e != '':
                    totest.append(e)
            if len(totest) == 0:
                totest = None
            else:
                totest = tuple(totest)
        assert col[idx] == totest, "Column: "+repr(col)+"\nSet value != gotten value "+str(totest)+" != "+str(col[idx])
        if totest is not None:
            for i in range(len(totest)) :
                assert col[idx][i] == totest[i], "Column: "+repr(col)+"\nSet value[i] != gotten value[i] "+str(totest[i])+" != "+str(col[idx][i])
    else:
        col[idx] = value
        assert col[idx] == value, "Column: "+repr(col)+"\nSet value != gotten value "+str(value)+" != "+str(col[idx])
    
    print_test(config, ">>> Set and get test OK")


def test_add_col(config, infos):
    print_test(config, ">>> Add column test")
    #existing_col = random_bool(config)    # TODO doesn't work because of line count problem. See obiview.c line 1737
    #if existing_col and infos["view_names"] != [] :
    #    random_view = infos['dms'].open_view(random.choice(infos["view_names"]))
    #    random_column = random_view[random.choice(sorted(list(random_view.columns))]
    #    random_column_refs = random_column.refs
    #    if random_column_refs['name'] in infos['view'] :
    #        alias = random_unique_name(infos)
    #    else :
    #        alias = ''
    #    infos['view'].add_column(random_column_refs['name'], version_number=random_column_refs['version'], alias=alias, create=False)
    #    random_view.close()
    #else :
    create_random_column(config, infos)
    print_test(config, ">>> Add column test OK")


def test_delete_col(config, infos):
    print_test(config, ">>> Delete column test")
    if len(list(infos['view'].keys())) <= 1 :
        print_test(config, "-")
        return
    col_name = random_column(infos)
    if col_name in SPECIAL_COLUMNS :
        print_test(config, "-")
        return
    infos['view'].delete_column(col_name)
    print_test(config, ">>> Delete column test OK")


def test_col_alias(config, infos):
    print_test(config, ">>> Changing column alias test")
    col_name = random_column(infos)
    if col_name in SPECIAL_COLUMNS :
        print_test(config, "-")
        return
    infos['view'][col_name].name = random_unique_name(infos)
    print_test(config, ">>> Changing column alias test OK")
         
 
def test_new_view(config, infos):
    print_test(config, ">>> New view test")
    random_new_view(config, infos)
    print_test(config, ">>> New view test OK")
 
 
def random_test(config, infos):
    return random.choice(infos['tests'])(config, infos)
 
 
def random_view_type():
    return random.choice(VIEW_TYPES)
 
 
def random_col_type():
    return random.choice(COL_TYPES)    
 
 
def fill_column(config, infos, col) :
    data_type = col.data_type
    element_names = col.elements_names
     
    if len(element_names) > 1 :
        for i in range(random_int(config)) :
            for j in range(len(element_names)) :
                col[i][element_names[j]] = random.choice([None, infos['random_generator'][(data_type, col.tuples)](config)])
    else :
        for i in range(random_int(config)) :
            r = random.choice([None, infos['random_generator'][(data_type, col.tuples)](config)])
            col[i] = r


def create_random_column(config, infos) :
    alias = random.choice([b'', random_unique_name(infos)])
    tuples = random.choice([True, False])
    dict_column = False
    if not tuples :
        nb_elements_per_line=random.randint(1, config['test']['maxelts'])
        if nb_elements_per_line > 1:
            dict_column = True
        elements_names = []
        for i in range(nb_elements_per_line) :
            elements_names.append(random_unique_element_name(config, infos))
        elements_names = random.choice([None, elements_names])
    else :
        nb_elements_per_line = 1
        elements_names = None
    name = random_unique_name(infos)
    data_type = random_col_type()
    
    column = Column.new_column(infos['view'], 
                               name, 
                               data_type, 
                               nb_elements_per_line=nb_elements_per_line,
                               elements_names=elements_names,
                               dict_column=dict_column,
                               tuples=tuples,
                               comments=random_comments(config),
                               alias=alias
                               )   
    
    if alias != b'' :
        assert infos['view'][alias] == column
    else :
        assert infos['view'][name] == column

    return column
 
 
def fill_view(config, infos):
    for i in range(random.randint(1, config['test']['maxinicolcount'])) :
        col = create_random_column(config, infos)
        fill_column(config, infos, col)
 
 
def random_new_view(config, infos, first=False):
    v_to_clone = None
    line_selection = None
    quality_col = False     # TODO
    if not first:
        infos['view_names'].append(infos['view'].name)
        infos['view'].close()
        v_to_clone = View.open(infos['dms'], random.choice(infos["view_names"]))
        v_type = b""
        print_test(config, "View to clone: ")
        print_test(config, repr(v_to_clone))
        create_line_selection = random_bool(config)
        if create_line_selection and v_to_clone.line_count > 0:
            print_test(config, "New view with new line selection.")
            line_selection = Line_selection(v_to_clone)
            for i in range(random.randint(1, v_to_clone.line_count)) :
                line_selection.append(random.randint(0, v_to_clone.line_count-1))
            #print_test(config, "New line selection: "+str(line_selection))
    else :
        v_type = random_view_type()
     
    if line_selection is not None :
        infos['view'] = line_selection.materialize(random_unique_name(infos), comments=random_comments(config))
    elif v_to_clone is not None :
        infos['view'] = v_to_clone.clone(random_unique_name(infos), comments=random_comments(config))
    else :
        if v_type == "NUC_SEQS_VIEW" :
            infos['view'] = View_NUC_SEQS.new(infos['dms'], random_unique_name(infos), comments=random_comments(config))   # TODO quality column
        else :
            infos['view'] = View.new(infos['dms'], random_unique_name(infos), comments=random_comments(config))   # TODO quality column
        infos['view'].write_config(config, "test", infos["command_line"], input_dms_name=[infos['dms'].name], input_view_name=["random"])
    print_test(config, repr(infos['view']))
    if v_to_clone is not None :
        if line_selection is None:
            assert v_to_clone.line_count == infos['view'].line_count, "New view and cloned view don't have the same line count : "+str(v_to_clone.line_count)+" (view to clone line count) != "+str(infos['view'].line_count)+" (new view line count)"
        else :
            assert len(line_selection) == infos['view'].line_count, "New view with new line selection does not have the right line count : "+str(len(line_selection))+" (line selection length) != "+str(infos['view'].line_count)+" (new view line count)"
        v_to_clone.close()
    if first :
        fill_view(config, infos)


def create_test_obidms(config, infos):
    infos['dms'] = DMS.new(config['obi']['defaultdms'])


def ini_dms_and_first_view(config, infos):
    create_test_obidms(config, infos)
    random_new_view(config, infos, first=True)
    infos['view_names'] = []


def addOptions(parser):
     
    # TODO put this common group somewhere else but I don't know where
    group=parser.add_argument_group('DMS and view options')
 
    group.add_argument('--default-dms','-d', 
                       action="store", dest="obi:defaultdms",
                       metavar='<DMS NAME>',
                       default="/tmp/test_dms",
                       type=str,
                       help="Name of the default DMS for reading and writing data. "
                            "Default: /tmp/test_dms")
    
    group.add_argument('--taxo','-t',     # TODO I don't understand why the option is not registered if it is not set
                       action="store", dest="obi:taxo",
                       metavar='<TAXDUMP PATH>',
                       default='',  # TODO not None because if it's None, the option is not entered in the option dictionary.
                       type=str,
                       help="Path to a taxdump to test the taxonomy.")

     
    group=parser.add_argument_group('obi test specific options')
 
    group.add_argument('--nb_tests','-n',
                       action="store", dest="test:nbtests",
                       metavar='<NB_TESTS>',
                       default=1000,
                       type=int,
                       help="Number of tests to carry out. "
                            "Default: 1000")
 
    group.add_argument('--seq_max_len','-s',
                       action="store", dest="test:seqmaxlen",
                       metavar='<SEQ_MAX_LEN>',
                       default=200,
                       type=int,
                       help="Maximum length of DNA sequences. "
                            "Default: 200")
 
    group.add_argument('--str_max_len','-r',
                       action="store", dest="test:strmaxlen",
                       metavar='<STR_MAX_LEN>',
                       default=200,
                       type=int,
                       help="Maximum length of character strings. "
                            "Default: 200")

    group.add_argument('--tuple_max_len','-u',
                       action="store", dest="test:tuplemaxlen",
                       metavar='<TUPLE_MAX_LEN>',
                       default=20,
                       type=int,
                       help="Maximum length of tuples. "
                            "Default: 20")
  
    group.add_argument('--max_ini_col_count','-o',
                       action="store", dest="test:maxinicolcount",
                       metavar='<MAX_INI_COL_COUNT>',
                       default=10,
                       type=int,
                       help="Maximum number of columns in the initial view. "
                            "Default: 10")
 
    group.add_argument('--max_line_nb','-l',
                       action="store", dest="test:maxlinenb",
                       metavar='<MAX_LINE_NB>',
                       default=1000,
                       type=int,
                       help="Maximum number of lines in a column. "
                            "Default: 1000")
 
    group.add_argument('--max_elts_per_line','-e',
                       action="store", dest="test:maxelts",
                       metavar='<MAX_ELTS_PER_LINE>',
                       default=20,
                       type=int,
                       help="Maximum number of elements per line in a column. "
                            "Default: 20")
 
    group.add_argument('--verbose','-v',
                       action="store_true", dest="test:verbose",
                       default=False,
                       help="Print the tests. "
                            "Default: Don't print the tests")
 
    group.add_argument('--seed','-g',
                       action="store", dest="test:seed",
                       metavar='<SEED>',
                       default=None,
                       help="Seed (use for reproducible tests). "
                            "Default: Seed is determined by Python")
 
def run(config):
     
    if 'seed' in config['test'] :
        random.seed(config['test']['seed'])
     
    infos = {'dms': None, 
             'view': None, 
             'view_names': None, 
             'unique_names': [],
             'random_generator': {
                                    (b"OBI_BOOL", False): random_bool, (b"OBI_BOOL", True): random_bool_tuples, 
                                    (b"OBI_CHAR", False): random_char, (b"OBI_CHAR", True): random_char_tuples, 
                                    (b"OBI_FLOAT", False): random_float, (b"OBI_FLOAT", True): random_float_tuples, 
                                    (b"OBI_INT", False): random_int, (b"OBI_INT", True): random_int_tuples, 
                                    (b"OBI_SEQ", False): random_seq, (b"OBI_SEQ", True): random_seq_tuples,
                                    (b"OBI_STR", False): random_bytes, (b"OBI_STR", True): random_bytes_tuples
                                  },
             'tests': [test_set_and_get, test_add_col, test_delete_col, test_col_alias, test_new_view],
             'command_line': " ".join(sys.argv[1:])
            }
 
    # TODO ???
    config['test']['elt_name_max_len'] = int((COL_COMMENTS_MAX_LEN - config['test']['maxelts']) / config['test']['maxelts'])
 
    print("Initializing the DMS and the first view...")
     
    shutil.rmtree(config['obi']['defaultdms']+'.obidms', ignore_errors=True)
    
    ini_dms_and_first_view(config, infos)
    print_test(config, repr(infos['view']))
    
    i = 0
    for t in range(config['test']['nbtests']):
        PyErr_CheckSignals()
        random_test(config, infos)
        print_test(config, repr(infos['view']))
        i+=1
        if (i%(config['test']['nbtests']/10)) == 0 :
            print("Testing......"+str(i*100/config['test']['nbtests'])+"%")
            #lsof = subprocess.Popen("lsof | grep '/private/tmp/test_dms.obidms/'", stdin=subprocess.PIPE, shell=True )
            #lsof.communicate( b"LSOF\n" )
            #lsof = subprocess.Popen("lsof | wc -l", stdin=subprocess.PIPE, shell=True )
            #lsof.communicate( b"LSOF total\n" )
 
    #print(infos)
    
    if config['obi']['taxo'] != '' :
        test_taxo(config, infos)
    
    infos['view'].close()
    infos['dms'].close(force=True)
    shutil.rmtree(config['obi']['defaultdms']+'.obidms', ignore_errors=True)
     
    print("Done.")

        


#cython: language_level=3

from obitools3.apps.progress cimport ProgressBar  # @UnresolvedImport
from obitools3.uri.decode import open_uri
from obitools3.apps.config import logger
from obitools3.dms import DMS
from obitools3.dms.obiseq import Nuc_Seq
from obitools3.dms.capi.obiview cimport QUALITY_COLUMN
from obitools3.writers.tab import TabWriter
from obitools3.format.tab import TabFormat
from obitools3.utils cimport tobytes, tostr

from obitools3.apps.optiongroups import addMinimalInputOption, \
                                        addExportOutputOption, \
                                        addNoProgressBarOption

import sys
import io

from cpython.exc cimport PyErr_CheckSignals

__title__="Export a view to a different file format"


def addOptions(parser):
    
    addMinimalInputOption(parser)
    addExportOutputOption(parser)
    addNoProgressBarOption(parser)


def run(config):
        
    DMS.obi_atexit()
    
    logger("info", "obi export : exports a view to a different file format")
    
    # Open the input
    input = open_uri(config['obi']['inputURI'])
    if input is None:
        raise Exception("Could not read input")
    iview = input[1]
    
    if 'outputformat' not in config['obi']:
        if iview.type == b"NUC_SEQS_VIEW":
            if QUALITY_COLUMN in iview:
                config['obi']['outputformat'] = b'fastq'
            else:
                config['obi']['outputformat'] = b'fasta'
        else:
            config['obi']['outputformat'] = b'tabular'
    
    # Open the output
    output = open_uri(config['obi']['outputURI'],
                      input=False)
    if output is None:
        raise Exception("Could not open output URI")

    output_object = output[0]
    writer = output[1]
        
     # Check that the input view has the type NUC_SEQS if needed    # TODO discuss, maybe bool property
    if (output[2] == Nuc_Seq) and (iview.type != b"NUC_SEQS_VIEW") :  # Nuc_Seq_Stored? TODO
        raise Exception("Error: the view to export in fasta or fastq format is not a NUC_SEQS view")
    
    if config['obi']['only'] is not None:
        withoutskip = min(input[4], config['obi']['only'])
    else:
        withoutskip = input[4]
     
    if config['obi']['skip'] is not None:
        skip = min(input[4], config['obi']['skip'])
    else:
        skip = 0
    
    # Initialize the progress bar
    if config['obi']['noprogressbar']:
        pb = None
    else:
        pb = ProgressBar(withoutskip - skip, config)

    if config['obi']['outputformat'] == b'metabaR':
        # Check prefix
        if "metabarprefix" not in config["obi"]:
            raise Exception("Prefix needed when exporting for metabaR (--metabaR-prefix option)")
        else:
            metabaRprefix = config["obi"]["metabarprefix"]
        
    i=0
    for seq in iview :
        PyErr_CheckSignals()
        if pb is not None:
            pb(i)
        try:
            writer(seq)
        except (StopIteration, BrokenPipeError, IOError):
            break
        i+=1

    if pb is not None:
        pb(i, force=True)
        print("", file=sys.stderr)

    if config['obi']['outputformat'] == b'metabaR':
        
        # Export ngsfilter file if view provided
        if 'metabarngsfilter' in config['obi']:
            ngsfilter_input = open_uri(config['obi']['metabarngsfilter'])
            if ngsfilter_input is None:
                raise Exception("Could not read ngsfilter view for metabaR output")
            ngsfilter_view = ngsfilter_input[1]
    
            ngsfilter_output = open(config['obi']['metabarprefix']+'.ngsfilter', 'w')
            
            for line in ngsfilter_view:

                line_to_print = b""                
                line_to_print += line[b'experiment']
                line_to_print += b"\t"
                line_to_print += line[b'sample']
                line_to_print += b"\t"
                line_to_print += line[b'forward_tag']
                line_to_print += b":"
                line_to_print += line[b'reverse_tag']
                line_to_print += b"\t"
                line_to_print += line[b'forward_primer']
                line_to_print += b"\t"
                line_to_print += line[b'reverse_primer']
                line_to_print += b"\t"
                line_to_print += line[b'additional_info']
                
                print(tostr(line_to_print), file=ngsfilter_output)
            
            if ngsfilter_input[0] != input[0]:
                ngsfilter_input[0].close()
            ngsfilter_output.close()
            
        # Export sample metadata
        samples_output = open(config['obi']['metabarprefix']+'_samples.csv', 'w')

        # Export sample metadata file if view provided
        if 'metabarsamples' in config['obi']:
            samples_input = open_uri(config['obi']['metabarsamples'])
            if samples_input is None:
                raise Exception("Could not read sample view for metabaR output")
            samples_view = samples_input[1]
        
            # Export with tab formatter
            TabWriter(TabFormat(header=True, sep='\t',), 
                      samples_output,
                      header=True)
            
            if samples_input[0] != input[0]:
                samples_input[0].close()
        
        # Else export just sample names from main view
        else:
            
            sample_list = []
            if 'MERGED_sample' in iview:
                sample_list = iview['MERGED_sample'].keys()
            elif 'sample' not in iview:
                for seq in iview:
                    sample = seq['sample']
                    if sample not in sample_list:
                        sample_list.append(sample)
            else:
                logger("warning", "Can not read sample list from main view for metabaR sample list export")            
            
            print("sample_id", file=samples_output)
            for sample in sample_list:
                line_to_print = b""                
                line_to_print += sample
                line_to_print += b"\t"
                print(tostr(line_to_print), file=samples_output)
        
        samples_output.close()

    # TODO save command in input dms?
    
    if not BrokenPipeError and not IOError:
        output_object.close()
    iview.close()
    input[0].close(force=True)

    logger("info", "Done.")

    if BrokenPipeError or IOError:
        sys.stderr.close()



#cython: language_level=3

from obitools3.dms.capi.obiview cimport Obiview_p
from obitools3.dms.capi.obidmscolumn cimport OBIDMS_column_p
from obitools3.dms.capi.obitypes cimport index_t

from libc.stdint  cimport int32_t, uint8_t


cdef extern from "kmer_similarity.h" nogil:

    struct Obi_ali_t :
        double   score
        int      consensus_length
        int      overlap_length
        char*    consensus_seq
        uint8_t* consensus_qual
        int      shift
        char*    direction

    ctypedef Obi_ali_t* Obi_ali_p

    
    void obi_free_shifted_ali(Obi_ali_p ali)


    Obi_ali_p kmer_similarity(Obiview_p view1,
                              OBIDMS_column_p column1,
                              index_t idx1,
                              index_t elt_idx1,
                              Obiview_p view2,
                              OBIDMS_column_p column2,
                              index_t idx2,
                              index_t elt_idx2,
                              OBIDMS_column_p qual_col1,
                              OBIDMS_column_p qual_col2,
                              OBIDMS_column_p reversed_col,
                              uint8_t kmer_size,
                              int32_t** kmer_pos_array,
                              int32_t* kmer_pos_array_height_p,
                              int32_t** shift_array,
                              int32_t* shift_array_height_p,
                              int32_t** shift_count_array, 
                              int32_t* shift_count_array_height_p,
                              bint build_consensus)



#cython: language_level=3

from obitools3.apps.progress cimport ProgressBar  # @UnresolvedImport
from obitools3.dms.dms cimport DMS
from obitools3.dms.view import RollbackException
from obitools3.dms.capi.obiclean cimport obi_clean
from obitools3.apps.optiongroups import addMinimalInputOption, addMinimalOutputOption, addNoProgressBarOption
from obitools3.uri.decode import open_uri
from obitools3.apps.config import logger
from obitools3.utils cimport tobytes, str2bytes
from obitools3.dms.view.view cimport View
from obitools3.dms.view.typed_view.view_NUC_SEQS cimport View_NUC_SEQS

from io import BufferedWriter
import sys


__title__="Tag a set of sequences for PCR and sequencing errors identification"


def addOptions(parser):

    addMinimalInputOption(parser)
    addMinimalOutputOption(parser)
    addNoProgressBarOption(parser)
    
    group = parser.add_argument_group('obi clean specific options')

    group.add_argument('--distance', '-d',
                       action="store", dest="clean:distance",
                       metavar='<DISTANCE>',
                       default=1.0,
                       type=float,
                       help="Maximum numbers of errors between two variant sequences. Default: 1.")

    group.add_argument('--sample-tag', '-s',
                       action="store", 
                       dest="clean:sample-tag-name",
                       metavar="<SAMPLE TAG NAME>",
                       type=str,
                       help="Name of the tag where merged sample count informations are kept (typically generated by obi uniq, usually MERGED_sample, default: None).")
    
    group.add_argument('--ratio', '-r',
                       action="store", dest="clean:ratio",
                       metavar='<RATIO>',
                       default=0.5,
                       type=float,
                       help="Maximum ratio between the counts of two sequences so that the less abundant one can be considered"
                            " a variant of the more abundant one. Default: 0.5.")
 
    group.add_argument('--heads-only', '-H',
                       action="store_true", 
                       dest="clean:heads-only",
                       default=False,
                       help="Only sequences labeled as heads are kept in the output. Default: False")

#    group.add_argument('--cluster-tags', '-C',
#                       action="store_true", 
#                       dest="clean:cluster-tags",
#                       default=False,
#                       help="Adds tags for each sequence giving its cluster's head and weight for each sample.")

    group.add_argument('--thread-count','-p',   # TODO should probably be in a specific option group
                       action="store", dest="clean:thread-count",
                       metavar='<THREAD COUNT>',
                       default=-1,
                       type=int,
                       help="Number of threads to use for the computation. Default: the maximum available.")


def run(config):
        
    DMS.obi_atexit()
    
    logger("info", "obi clean")

    # Open the input: only the DMS
    input = open_uri(config['obi']['inputURI'],
                     dms_only=True)
    if input is None:
        raise Exception("Could not read input")
    i_dms = input[0]
    i_dms_name = input[0].name
    i_view_name = input[1]

    # Open the output: only the DMS
    output = open_uri(config['obi']['outputURI'],
                      input=False,
                      dms_only=True)
    if output is None:
        raise Exception("Could not create output")
    o_dms = output[0]
    output_0 = output[0]
    final_o_view_name = output[1]
    
    # If stdout output or the input and output DMS are not the same, create a temporary view that will be exported to 
    # the right DMS and deleted in the other afterwards.
    if i_dms != o_dms or type(output_0)==BufferedWriter:
        temporary_view_name = b"temp"
        i=0
        while temporary_view_name in i_dms:  # Making sure view name is unique in input DMS
            temporary_view_name = temporary_view_name+b"_"+str2bytes(str(i))
            i+=1
        o_view_name = temporary_view_name
        if type(output_0)==BufferedWriter:
            o_dms = i_dms
    else:
        o_view_name = final_o_view_name
        
    # Save command config in View comments
    command_line = " ".join(sys.argv[1:])
    comments = View.print_config(config, "clean", command_line, input_dms_name=[i_dms_name], input_view_name=[i_view_name])

    if 'sample-tag-name' not in config['clean']:
        config['clean']['sample-tag-name'] = ""
        
    if obi_clean(i_dms.name_with_full_path, tobytes(i_view_name), tobytes(config['clean']['sample-tag-name']), tobytes(o_view_name), comments, \
              config['clean']['distance'], config['clean']['ratio'], config['clean']['heads-only'], config['clean']['thread-count']) < 0:
        raise Exception("Error running obiclean")
    
    # If the input and output DMS are not the same, export result view to output DMS
    if i_dms != o_dms:
        View.import_view(i_dms.full_path[:-7], o_dms.full_path[:-7], o_view_name, final_o_view_name)

    # stdout output: write to buffer
    if type(output_0)==BufferedWriter:
        logger("info", "Printing to output...")
        o_view = o_dms[o_view_name]
        o_view.print_to_output(output_0, noprogressbar=config['obi']['noprogressbar'])
        o_view.close()

    # Save command config in DMS comments
    o_dms.record_command_line(command_line)

    #print("\n\nOutput view:\n````````````", file=sys.stderr)
    #print(repr(o_dms[final_o_view_name]), file=sys.stderr)

    # If the input and the output DMS are different or if stdout output, delete the temporary imported view used to create the final view
    if i_dms != o_dms or type(output_0)==BufferedWriter:
        View.delete_view(i_dms, o_view_name)
        o_dms.close(force=True)
    
    i_dms.close(force=True)

    logger("info", "Done.")
    


#cython: language_level=3

from obitools3.apps.progress cimport ProgressBar  # @UnresolvedImport
from obitools3.dms import DMS
from obitools3.dms.view.view cimport View
from obitools3.uri.decode import open_uri
from obitools3.apps.optiongroups import addMinimalInputOption, addMinimalOutputOption, addNoProgressBarOption
from obitools3.dms.view import RollbackException
from obitools3.apps.config import logger
from obitools3.utils cimport tobytes, str2bytes

from obitools3.dms.capi.obilcsalign cimport obi_lcs_align_one_column, \
                                            obi_lcs_align_two_columns

from io import BufferedWriter
from cpython.exc cimport PyErr_CheckSignals

import time
import sys


__title__="Align one sequence column with itself or two sequence columns"


def addOptions(parser):
   
   addMinimalInputOption(parser)
   addMinimalOutputOption(parser)
   addNoProgressBarOption(parser)

   group=parser.add_argument_group('obi align specific options')

   group.add_argument('--input-2', '-I',
                      action="store", dest="align:inputuri2",
                      metavar='<INPUT URI>',
                      default="",
                      type=str,
                      help="Eventually, the URI of the second input to align with the first one.")
      
   group.add_argument('--threshold','-t',
                      action="store", dest="align:threshold",
                      metavar='<THRESHOLD>',
                      default=0.0,
                      type=float,
                      help="Score threshold. If the score is normalized and expressed in similarity (default),"
                           " it is an identity, e.g. 0.95 for an identity of 95%%. If the score is normalized"
                           " and expressed in distance, it is (1.0 - identity), e.g. 0.05 for an identity of 95%%."
                           " If the score is not normalized and expressed in similarity, it is the length of the"
                           " Longest Common Subsequence. If the score is not normalized and expressed in distance,"
                           " it is (reference length - LCS length)."
                           " Only sequence pairs with a similarity above <THRESHOLD> are printed. Default: 0.00"
                           " (no threshold).")
   
   group.add_argument('--longest-length','-L',
                      action="store_const", dest="align:reflength",
                      default=0,
                      const=1,
                      help="The reference length is the length of the longest sequence."
                           " Default: the reference length is the length of the alignment.")
   
   group.add_argument('--shortest-length','-l',
                      action="store_const", dest="align:reflength",
                      default=0,
                      const=2,
                      help="The reference length is the length of the shortest sequence."
                           " Default: the reference length is the length of the alignment.")
   
   group.add_argument('--raw','-r',
                      action="store_false", dest="align:normalize",
                      default=True,
                      help="Raw score, not normalized. Default: score is normalized with the reference sequence length.")
   
   group.add_argument('--distance','-D',
                      action="store_false", dest="align:similarity",
                      default=True,
                      help="Score is expressed in distance. Default: score is expressed in similarity.")
   
   group.add_argument('--print-seq','-s',
                      action="store_true", dest="align:printseq",
                      default=False,
                      help="The nucleotide sequences are written in the output view. Default: they are not written.")
   
   group.add_argument('--print-count','-n',
                      action="store_true", dest="align:printcount",
                      default=False,
                      help="Sequence counts are written in the output view. Default: they are not written.")
   
   group.add_argument('--thread-count','-p',   # TODO should probably be in a specific option group
                      action="store", dest="align:threadcount",
                      metavar='<THREAD COUNT>',
                      default=1,
                      type=int,
                      help="Number of threads to use for the computation. Default: one.")


cpdef align_columns(bytes dms_n, 
                    bytes input_view_1_n, 
                    bytes output_view_n,
                    bytes input_view_2_n=b"",
                    bytes input_column_1_n=b"", 
                    bytes input_column_2_n=b"",
                    bytes input_elt_1_n=b"", 
                    bytes input_elt_2_n=b"",
                    bytes id_column_1_n=b"", 
                    bytes id_column_2_n=b"",
                    double threshold=0.0, bint normalize=True, 
                    int reference=0, bint similarity_mode=True,
                    bint print_seq=False, bint print_count=False,
                    bytes comments=b"{}",
                    int thread_count=1) : 

    if input_view_2_n == b"" and input_column_2_n == b"" :
        if obi_lcs_align_one_column(dms_n, \
                                    input_view_1_n, \
                                    input_column_1_n, \
                                    input_elt_1_n, \
                                    id_column_1_n, \
                                    output_view_n, \
                                    comments, \
                                    print_seq, \
                                    print_count, \
                                    threshold, normalize, reference, similarity_mode,
                                    thread_count) < 0 :
            raise Exception("Error aligning sequences")        
            
    else:
        if obi_lcs_align_two_columns(dms_n, \
                                     input_view_1_n, \
                                     input_view_2_n, \
                                     input_column_1_n, \
                                     input_column_2_n, \
                                     input_elt_1_n, \
                                     input_elt_2_n, \
                                     id_column_1_n, \
                                     id_column_2_n, \
                                     output_view_n, \
                                     comments, \
                                     print_seq, \
                                     print_count, \
                                     threshold, normalize, reference, similarity_mode) < 0 :
            raise Exception("Error aligning sequences")        
 
 
def run(config):
    
    DMS.obi_atexit()
    
    logger("info", "obi align")

    # Open the input: only the DMS
    input = open_uri(config['obi']['inputURI'],
                     dms_only=True)
    if input is None:
        raise Exception("Could not read input")
    i_dms = input[0]
    i_dms_name = input[0].name
    i_uri = input[1]
    i_view_name = i_uri.split(b"/")[0]
    i_column_name = b""
    i_element_name = b""
    if len(i_uri.split(b"/")) >= 2:
        i_column_name = i_uri.split(b"/")[1]
    if len(i_uri.split(b"/")) == 3:
        i_element_name = i_uri.split(b"/")[2]
    if len(i_uri.split(b"/")) > 3:
        raise Exception("Input URI contains too many elements:", config['obi']['inputURI'])

    # Open the second input if there is one
    i_dms_2 = None
    i_dms_name_2 = b""
    original_i_view_name_2 = b""
    i_view_name_2 = b""
    i_column_name_2 = b""
    i_element_name_2 = b""
    if config['align']['inputuri2']:
        input_2 = open_uri(config['align']['inputuri2'],
                           dms_only=True)
        if input_2 is None:
            raise Exception("Could not read second input")
        i_dms_2 = input_2[0]
        i_dms_name_2 = i_dms_2.name
        i_uri_2 = input_2[1]
        original_i_view_name_2 = i_uri_2.split(b"/")[0]
        if len(i_uri_2.split(b"/")) >= 2:
            i_column_name_2 = i_uri_2.split(b"/")[1]
        if len(i_uri_2.split(b"/")) == 3:
            i_element_name_2 = i_uri_2.split(b"/")[2]
        if len(i_uri_2.split(b"/")) > 3:
            raise Exception("Input URI contains too many elements:", config['align']['inputuri2'])

        # If the 2 input DMS are not the same, temporarily import 2nd input view in first input DMS
        if i_dms != i_dms_2:
            temp_i_view_name_2 = original_i_view_name_2
            i=0
            while temp_i_view_name_2 in i_dms:  # Making sure view name is unique in input DMS
                temp_i_view_name_2 = original_i_view_name_2+b"_"+str2bytes(str(i))
                i+=1
            i_view_name_2 = temp_i_view_name_2
            View.import_view(i_dms_2.full_path[:-7], i_dms.full_path[:-7], original_i_view_name_2, i_view_name_2)

    # Open the output: only the DMS
    output = open_uri(config['obi']['outputURI'],
                      input=False,
                      dms_only=True)
    if output is None:
        raise Exception("Could not create output")
    o_dms = output[0]
    output_0 = output[0]
    o_dms_name = o_dms.name 
    final_o_view_name = output[1]
    o_view_name = final_o_view_name

    # If stdout output or the input and output DMS are not the same, align creating a temporary view in the input dms that will be exported to 
    # the right DMS and deleted in the other afterwards.
    if i_dms != o_dms or type(output_0)==BufferedWriter:
        if type(output_0)==BufferedWriter:
            o_dms = i_dms
        o_view_name = b"temp"
        while o_view_name in i_dms:   # Making sure view name is unique in input DMS
            o_view_name = o_view_name+b"_"+str2bytes(str(i))
            i+=1
        
    # Save command config in View comments
    command_line = " ".join(sys.argv[1:])
    
    i_dms_list = [i_dms_name]
    if i_dms_name_2:
        i_dms_list.append(i_dms_name_2)
    i_view_list = [i_view_name]
    if original_i_view_name_2:
        i_view_list.append(original_i_view_name_2)
    comments = View.print_config(config, "align", command_line, input_dms_name=i_dms_list, input_view_name=i_view_list)
    
    # Call cython alignment function
      # Using default ID columns of the view. TODO discuss adding option
    align_columns(i_dms.name_with_full_path,  \
                  i_view_name,  \
                  o_view_name,  \
                  input_view_2_n   = i_view_name_2,  \
                  input_column_1_n = i_column_name,  \
                  input_column_2_n = i_column_name_2, \
                  input_elt_1_n    = i_element_name,  \
                  input_elt_2_n    = i_element_name_2, \
                  id_column_1_n    = b"",  \
                  id_column_2_n    = b"", \
                  threshold        = config['align']['threshold'], \
                  normalize        = config['align']['normalize'],  \
                  reference        = config['align']['reflength'],  \
                  similarity_mode  = config['align']['similarity'],  \
                  print_seq        = config['align']['printseq'],  \
                  print_count      = config['align']['printcount'], \
                  comments         = comments, \
                  thread_count     = config['align']['threadcount'])

    # If the input and output DMS are not the same, export result view to output DMS
    if i_dms != o_dms:
        View.import_view(i_dms.full_path[:-7], o_dms.full_path[:-7], o_view_name, final_o_view_name)

    # Save command config in output DMS comments
    o_dms.record_command_line(command_line)

    #print("\n\nOutput view:\n````````````", file=sys.stderr)
    #print(repr(o_dms[final_o_view_name]), file=sys.stderr)

    # If the two input DMS are different, delete the temporary input view in the first input DMS
    if i_dms_2 and i_dms != i_dms_2:
        View.delete_view(i_dms, i_view_name_2)
        i_dms_2.close()

    # stdout output: write to buffer
    if type(output_0)==BufferedWriter:
        logger("info", "Printing to output...")
        o_view = o_dms[o_view_name]
        o_view.print_to_output(output_0, noprogressbar=config['obi']['noprogressbar'])
        o_view.close()

    # If the input and the output DMS are different, delete the temporary result view in the input DMS
    if i_dms != o_dms or type(output_0)==BufferedWriter:
        View.delete_view(i_dms, o_view_name)
        o_dms.close(force=True)
    
    i_dms.close(force=True)
    
    logger("info", "Done.")
  


#cython: language_level=3

'''
Created on oct 12th 2018

@author: celine.mercier.bioinfo@gmail.com
'''

    
cdef class FastaNucWriter:
    
    def __init__(self, 
                 object formatter, 
                 object output_object, 
                 int skip=0,
                 only=None):
                        
        if only is None:
            self.only = -1
        else:
            self.only = int(only)

        self.formatter = formatter
        self.output = output_object
        self.skip = skip
        self.skipped = 0
        self.read = 0
    
    def __call__(self, object seq):
        if self.only > -1 and self.read == self.only:
            raise StopIteration
        if self.skip > 0 and self.skipped < self.skip:
            self.skipped += 1
            return
        self.output.write(self.formatter(seq))
        self.output.write(b"\n")  # TODO is that clean?
        self.read += 1


#cython: language_level=3

from obitools3.apps.progress cimport ProgressBar  # @UnresolvedImport
from obitools3.dms import DMS
from obitools3.dms.view.typed_view.view_NUC_SEQS cimport View_NUC_SEQS
from obitools3.dms.column.column cimport Column
from obitools3.dms.capi.obiview cimport QUALITY_COLUMN
from obitools3.dms.capi.obitypes cimport OBI_QUAL
from obitools3.apps.optiongroups import addMinimalInputOption, addMinimalOutputOption, addNoProgressBarOption
from obitools3.uri.decode import open_uri
from obitools3.apps.config import logger
from obitools3.libalign._qsassemble import QSolexaReverseAssemble
from obitools3.libalign._qsrassemble import QSolexaRightReverseAssemble
from obitools3.libalign._solexapairend import buildConsensus, buildJoinedSequence
from obitools3.dms.obiseq cimport Nuc_Seq
from obitools3.libalign.shifted_ali cimport Kmer_similarity, Ali_shifted
from obitools3.dms.capi.obiview cimport REVERSE_SEQUENCE_COLUMN, REVERSE_QUALITY_COLUMN
from obitools3.utils cimport str2bytes

from io import BufferedWriter
import sys
import os

from cpython.exc cimport PyErr_CheckSignals

__title__="Align paired-ended reads"



def addOptions(parser):

    addMinimalInputOption(parser)
    addMinimalOutputOption(parser)
    addNoProgressBarOption(parser)
 
    group = parser.add_argument_group('obi alignpairedend specific options')

    group.add_argument('-R', '--reverse-reads',
                     action="store", dest="alignpairedend:reverse",
                     metavar="<URI>",
                     default=None,
                     type=str,
                     help="URI to the reverse reads if they are in a different view than the forward reads")

#     group.add_argument('--score-min',
#                      action="store", dest="alignpairedend:smin",
#                      metavar="#.###",
#                      default=None,
#                      type=float,
#                      help="Minimum score for keeping alignments. " 
#                           "(for kmer alignment) The score is an approximation of the number of nucleotides matching in the overlap of the alignment.")

#     group.add_argument('-A', '--true-ali',
#                        action="store_true", dest="alignpairedend:trueali",
#                        default=False,
#                        help="Performs gap free end alignment of sequences instead of using kmers to compute alignments (slower).")

    group.add_argument('-k', '--kmer-size',
                       action="store", dest="alignpairedend:kmersize",
                       metavar="#",
                       default=3,
                       type=int,
                       help="K-mer size for kmer comparisons, between 1 and 4 (default: 3)")


la = QSolexaReverseAssemble()
ra = QSolexaRightReverseAssemble()
cdef object buildAlignment(object direct, object reverse):
    
    if len(direct)==0 or len(reverse)==0:
        return None
    
    la.seqA = direct
    la.seqB = reverse
    
    ali=la()
    ali.direction='left'

    ra.seqA = direct
    ra.seqB = reverse

    rali=ra()
    rali.direction='right'
     
    if ali.score < rali.score:
        ali = rali
    
    return ali
    

def alignmentIterator(entries, aligner): 
    
    if type(entries) == list:
        two_views = True
        forward = entries[0]
        reverse = entries[1]
        entries_len = len(forward)
    else:
        two_views = False
        entries_len = len(entries)
    
    for i in range(entries_len):
                
        if two_views:
            seqF = forward[i]
            seqR = reverse[i]
        else:
            seqF = Nuc_Seq.new_from_stored(entries[i])
            seqR = Nuc_Seq(seqF.id, seqF[REVERSE_SEQUENCE_COLUMN], quality=seqF[REVERSE_QUALITY_COLUMN])
            seqR.index = i
        
        ali = aligner(seqF, seqR)
        
        if ali is None:
            continue
        
        yield ali


def run(config):
        
    DMS.obi_atexit()
    
    logger("info", "obi alignpairedend")
       
    # Open the input
    
    two_views = False
    forward = None
    reverse = None
    input = None
    
    input = open_uri(config['obi']['inputURI'])
    if input is None:
        raise Exception("Could not open input reads")
    if input[2] != View_NUC_SEQS:
        raise NotImplementedError('obi alignpairedend only works on NUC_SEQS views')    
    
    if "reverse" in config["alignpairedend"]:
        
        two_views = True
        
        forward = input[1]        
        
        rinput = open_uri(config["alignpairedend"]["reverse"])
        if rinput is None:
            raise Exception("Could not open reverse reads")
        if rinput[2] != View_NUC_SEQS:
            raise NotImplementedError('obi alignpairedend only works on NUC_SEQS views')
        
        reverse = rinput[1]
    
        if len(forward) != len(reverse):
            raise Exception("Error: the number of forward and reverse reads are different")
        
        entries = [forward, reverse]
        input_dms_name = [forward.dms.name, reverse.dms.name]
        input_view_name = [forward.name, reverse.name]
    
    else:
        entries = input[1]
        input_dms_name = [entries.dms.name]
        input_view_name = [entries.name]

    if two_views:
        entries_len = len(forward)
    else:
        entries_len = len(entries)
    
    # Open the output
    output = open_uri(config['obi']['outputURI'],
                      input=False,
                      newviewtype=View_NUC_SEQS)
    if output is None:
        raise Exception("Could not create output view")

    output_0 = output[0]
    o_dms = output[0]
    
    # stdout output: create temporary view
    if type(output_0)==BufferedWriter:
        i_dms = forward.dms   # using any dms
        o_dms = i_dms
        i=0
        o_view_name = b"temp"
        while o_view_name in i_dms:   # Making sure view name is unique in input DMS
            o_view_name = o_view_name+b"_"+str2bytes(str(i))
            i+=1
        o_view = View_NUC_SEQS.new(o_dms, o_view_name, quality=True)
    else:
        o_view = output[1]
        Column.new_column(o_view, QUALITY_COLUMN, OBI_QUAL)

    # Initialize the progress bar
    if config['obi']['noprogressbar'] == False:
        pb = ProgressBar(entries_len, config)
    else:
        pb = None
         
    #if config['alignpairedend']['trueali']:
    #    kmer_ali = False
    #    aligner = buildAlignment
    #else :
    kmer_ali = True
    if type(entries) == list:
        forward = entries[0]
        reverse = entries[1]
        if len(forward) == 0 or len(reverse) == 0:
            aligner = None
        else:
            aligner = Kmer_similarity(forward, \
                                      view2=reverse, \
                                      kmer_size=config['alignpairedend']['kmersize'], \
                                      reversed_column=None)
    else:
        if len(entries) == 0:
            aligner = None
        else:
            aligner = Kmer_similarity(entries, \
                                      column2=entries[REVERSE_SEQUENCE_COLUMN], \
                                      qual_column2=entries[REVERSE_QUALITY_COLUMN], \
                                      kmer_size=config['alignpairedend']['kmersize'], \
                                      reversed_column=entries[b'reversed'])  # column created by the ngsfilter tool
        
    ba = alignmentIterator(entries, aligner)
    
    i = 0
    for ali in ba:
        
        if pb is not None:
            pb(i)
        
        PyErr_CheckSignals()

        consensus = o_view[i]
        
        if two_views:
            consensus[b"R1_parent"] = forward[i].id
            consensus[b"R2_parent"] = reverse[i].id
               
        if not two_views:
            seqF = entries[i]
        else:
            seqF = forward[i]
                        
        if ali.overlap_len > 0 :
            buildConsensus(ali, consensus, seqF)
        else:
            if not two_views:
                seqR = Nuc_Seq(seqF.id, seqF[REVERSE_SEQUENCE_COLUMN], quality = seqF[REVERSE_QUALITY_COLUMN])
            else:
                seqR = reverse[i]
            buildJoinedSequence(ali, seqR, consensus, forward=seqF)
                   
        if kmer_ali :
            ali.free()
    
        i+=1

    if pb is not None:
        pb(i, force=True)
        print("", file=sys.stderr)

    if kmer_ali and aligner is not None:
        aligner.free()

    # Save command config in View and DMS comments
    command_line = " ".join(sys.argv[1:])
    o_view.write_config(config, "alignpairedend", command_line, input_dms_name=input_dms_name, input_view_name=input_view_name)
    o_dms.record_command_line(command_line)

    #print("\n\nOutput view:\n````````````", file=sys.stderr)
    #print(repr(view), file=sys.stderr)

    # stdout output: write to buffer
    if type(output_0)==BufferedWriter:
        logger("info", "Printing to output...")
        o_view.print_to_output(output_0, noprogressbar=config['obi']['noprogressbar'])
        o_view.close()

    # If stdout output, delete the temporary imported view used to create the final file
    if type(output_0)==BufferedWriter:
        View_NUC_SEQS.delete_view(o_dms, o_view_name)
        output_0.close()

    # Close all DMS
    input[0].close(force=True)
    if two_views:
        rinput[0].close(force=True)
    o_dms.close(force=True)

    logger("info", "Done.")
    


#cython: language_level=3

from obitools3.dms.capi.obidms       cimport OBIDMS_p


cdef extern from "obi_clean.h" nogil:

    int obi_clean(const char* dms_name,
                  const char* i_view_name,
                  const char* sample_column_name,
                  const char* o_view_name,
                  const char* o_view_comments,
                  double threshold,
                  double max_ratio,
                  bint heads_only,
                  int thread_count)
    
